
# File: solutiondescription.py 
# Author(s): Saswati Ray
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

import pandas as pd
import uuid, copy, math
import time, random
import numpy as np
import json
from d3m.metadata.pipeline import Pipeline, PrimitiveStep, SubpipelineStep
from d3m.metadata.base import PrimitiveFamily, Context, ArgumentType
from d3m.metadata import base as metadata_base
from d3m.primitive_interfaces.base import PrimitiveBaseMeta
from d3m.runtime import Runtime
from d3m import container
import d3m.index

import networkx as nx

import logging, typing, os 
from timeit import default_timer as timer

# AutonML imports
from autonml.helper_functions import StepType as StepType
from autonml.helper_functions import ActionType as ActionType
from autonml.helper_functions import get_argument as get_argument
from autonml.helper_functions import get_pipeline_values as get_pipeline_values
from autonml.helper_functions import get_values as get_values
from autonml.helper_functions import get_split_indices, get_num_splits
import autonml.solution_templates as solution_templates 
import autonml.default_hyperparams as default_hyperparams
import autonml.helper_functions as helper_functions
import autonml.problem_pb2 as problem_pb2
import autonml.pipeline_pb2 as pipeline_pb2 
import autonml.primitive_pb2 as primitive_pb2
import autonml.value_pb2 as value_pb2

class SolutionDescription(object):
    """
    Represents a solution or pipeline in the D3M environment.
    The idea is that this can be evaluated, fit, produce a model and performance metrics.
    """
    def __init__(self, problem):
        self.id = str(uuid.uuid4())
        self.source = None
        self.created = helper_functions.compute_timestamp()
        self.context = 'PRETRAINING'
        self.name = None
        self.description = None
        self.users = None
        self.inputs = []
        self.rank = 1
        
        self.problem = problem
        self.outputs = None
        self.execution_order = None
        self.primitives_arguments = None
        self.primitives = None
        self.subpipelines = None
        self.pipeline = None
        self.produce_order = None
        self.hyperparams = None
        self.steptypes = None
        self.taskname = None
        self.exclude_columns = None
        self.primitives_outputs = None
        self.cross_outputs = None
        self.pipeline_description = None
        self.total_cols = 0
        self.categorical_atts = None
        self.ordinal_atts = None
        self.ok_to_denormalize = True
        self.add_floats = None
        self.privileged = None
        self.index_denormalize = 0
        self.pipeline_description = None
        self.splitter = False
        self.volumes_dir = os.environ['D3MSTATICDIR']

    def set_categorical_atts(self, atts):
        self.categorical_atts = atts

    def set_ordinal_atts(self, atts):
        self.ordinal_atts = atts

    def set_denormalize(self, ok_to_denormalize):
        self.ok_to_denormalize = ok_to_denormalize

    def set_privileged(self, privileged):
        self.privileged = privileged

    def set_add_floats(self, add_floats):
        self.add_floats = add_floats

    def add_splitter(self):
        self.splitter = True

    def contains_placeholder(self):
        if self.steptypes is None:
            return False

        for step in self.steptypes:
            if step == StepType.PLACEHOLDER:
                return True
        return False

    def num_steps(self):
        """
        Returns number of steps in the pipeline.
        """
        if self.primitives_arguments is not None:
            return len(self.primitives_arguments)
        else:
            return 0

    def find_primitive_index(self, path):
        """
        Returns index of particular primitive in the pipeline
        """
        index = 0
        for id,p in self.primitives.items():
            python_path = p.metadata.query()['python_path'] 
            if path == python_path:
                return index
            index = index + 1

    def create_pipeline_json(self, prim_dict, solution_dict):
        """
        Generate pipeline.json
        """
        name = "Pipeline for evaluation"
        pipeline_id = self.id 
        pipeline_description = Pipeline(pipeline_id=pipeline_id, name=name)
        for ip in self.inputs:
            pipeline_description.add_input(name=ip['name'])

        num = self.num_steps()
 
        # Iterate through different steps in the pipeline
        for i in range(num):
            if self.steptypes[i] == StepType.PRIMITIVE: # PRIMITIVE
                p = prim_dict[self.primitives[i]]
                pdesc = {}
                pdesc['id'] = p.id
                pdesc['version'] = p.primitive_class.version
                pdesc['python_path'] = p.primitive_class.python_path
                pdesc['name'] = p.primitive_class.name
                pdesc['digest'] = p.primitive_class.digest
                step = PrimitiveStep(primitive_description=pdesc)

                # Primitives with multiple produce methods being used in the pipeline
                if 'DistilSingleGraphLoader' in p.primitive_class.python_path or \
                   'DistilGraphLoader' in p.primitive_class.python_path or \
                   'DistilEdgeListLoader' in p.primitive_class.python_path or \
                   'DistilAudioDatasetLoader' in p.primitive_class.python_path:
                    for op in p.primitive_class.produce_methods:
                        step.add_output(output_id=op)
                elif 'extra_trees' in p.primitive_class.python_path or \
                     'random_forest' in p.primitive_class.python_path or \
                     'gradient_boosting' in p.primitive_class.python_path or \
                     'ada_boost' in p.primitive_class.python_path:
                     #'xgboost_gbtree' in p.primitive_class.python_path:
                    for op in p.primitive_class.produce_methods:
                        step.add_output(output_id=op)
                else: # Single produce methods
                    if len(self.primitives_arguments[i]) > 0:
                        step.add_output(output_id=p.primitive_class.produce_methods[0])
                for name, value in self.primitives_arguments[i].items():
                    origin = value['origin']
                    argument_type = ArgumentType.CONTAINER
                    step.add_argument(name=name, argument_type=argument_type, data_reference=value['data'])
                if self.hyperparams[i] is not None:
                    for name, value in self.hyperparams[i].items():
                        if name == "primitive" or name == "primitive_learner":
                            step.add_hyperparameter(name=name, argument_type=ArgumentType.PRIMITIVE, data=value) 
                        else:
                            step.add_hyperparameter(name=name, argument_type=ArgumentType.VALUE, data=value)
            else: # SUBPIPELINE
                solution = solution_dict[self.subpipelines[i]]
                pdesc = solution.pipeline_description
                if pdesc is None:
                    solution.create_pipeline_json(prim_dict, solution_dict)
                    pdesc = solution.pipeline_description 
                step = SubpipelineStep(pipeline=pdesc)
                ipname = 'steps.' + str(i-1) + '.produce'
                step.add_input(ipname)
                for output in solution.outputs:
                    step.add_output(output_id=output[2])

            pipeline_description.add_step(step)

        for op in self.outputs:
            pipeline_description.add_output(data_reference=op[2], name=op[3])

        self.pipeline_description = pipeline_description
   
    def write_pipeline_json(self, prim_dict, solution_dict, dirName, subpipeline_dirName, rank=None, train_score=None):
        """
        Output pipeline to JSON file
        """
        filename = dirName + "/" + self.id + ".json"
        if self.pipeline_description is None:
            self.create_pipeline_json(prim_dict, solution_dict)
            for step in self.pipeline_description.steps:
                if isinstance(step, SubpipelineStep):
                    subfilename = subpipeline_dirName + "/" + step.pipeline.id + ".json"
                    with open(subfilename, "w") as out:
                        parsed = json.loads(step.pipeline.to_json())
                        json.dump(parsed, out, indent=4)

        with open(filename, "w") as out:
            parsed = json.loads(self.pipeline_description.to_json())
            if rank is not None:
                parsed['pipeline_rank'] = str(rank)
            if rank is not None:
                parsed['pipeline_score'] = train_score
            json.dump(parsed, out, indent=4)    

    def write_pipeline_run(self, problem_description, dataset, filename_yaml): 
        runtime = Runtime(pipeline=self.pipeline_description, problem_description=problem_description, context=Context.TESTING, is_standard_pipeline=True)
        output = runtime.fit(inputs=dataset)
        pipeline_run = output.pipeline_run
        with open(filename_yaml, "w") as out:
            pipeline_run.to_yaml(file=filename_yaml)

    def create_from_pipelinedescription(self, solution_dict, pipeline_description: pipeline_pb2.PipelineDescription) -> None:
        """
        Initialize a solution object from a pipeline_pb2.PipelineDescription object passed by TA3.
        """
        n_steps = len(pipeline_description.steps)

        logging.critical("Steps = %d", n_steps)
        logging.critical("Running %s", pipeline_description)
       
        self.inputs = []
        for ip in pipeline_description.inputs:
             self.inputs.append({"name": ip.name})

        self.id = pipeline_description.id
        self.source = {}
        self.source['name'] = pipeline_description.source.name
        self.source['contact'] = pipeline_description.source.contact
        self.source['pipelines'] = []
        for id in pipeline_description.source.pipelines:
            self.source['pipelines'].append(id)

        self.name = pipeline_description.name
        self.description = pipeline_description.description
        self.users = []
        for user in pipeline_description.users:
            self.users.append({"id": user.id, "reason": user.reason, "rationale": user.rationale})

        self.subpipelines = {}
        self.primitives_arguments = {}
        self.primitives = {}
        self.hyperparams = {}
        self.steptypes = []
        for i in range(0, n_steps):
            self.primitives_arguments[i] = {}
            self.hyperparams[i] = None

        self.execution_order = None

        self.pipeline = [None] * n_steps
        self.outputs = []

        # Constructing DAG to determine the execution order
        execution_graph = nx.DiGraph()
        for i in range(0, n_steps):
            # PrimitivePipelineDescriptionStep
            if pipeline_description.steps[i].HasField("primitive") == True:
                s = pipeline_description.steps[i].primitive
                python_path = s.primitive.python_path
                prim = d3m.index.get_primitive(python_path)
                self.primitives[i] = prim
                self.subpipelines[i] = None
                arguments = s.arguments
                self.steptypes.append(StepType.PRIMITIVE)

                for name,argument in arguments.items():
                    if argument.HasField("container") == True:
                        data = argument.container.data
                    else:
                        data = argument.data.data
                    origin = data.split('.')[0]
                    source = data.split('.')[1]
                    self.primitives_arguments[i][name] = {'origin': origin, 'source': int(source), 'data': data}
                    
                    if origin == 'steps':
                        execution_graph.add_edge(str(source), str(i))
                    else:
                        execution_graph.add_edge(origin, str(i))

                hyperparams = s.hyperparams
                if hyperparams is not None:
                    self.hyperparams[i] = {}
                    for name,argument in hyperparams.items():
                        value = None
                        if argument.HasField("value") == True:
                            arg = argument.value.data.raw
                            value = get_values(arg)
                            self.hyperparams[i][name] = value
                        elif argument.HasField("primitive") == True:
                            arg = int(argument.primitive.data)
                            self.hyperparams[i][name] = arg 

            # SubpipelinePipelineDescriptionStep
            elif pipeline_description.steps[i].HasField("pipeline") == True:
                s = pipeline_description.steps[i].pipeline
                self.primitives[i] = None
                self.subpipelines[i] = s.pipeline.id
                self.steptypes.append(StepType.SUBPIPELINE)
                for j in range(len(s.inputs)):
                    argument_edge = s.inputs[j].data
                    origin = argument_edge.split('.')[0]
                    source = argument_edge.split('.')[1]
                    self.primitives_arguments[i]['inputs'] = {'origin': origin, 'source': int(source), 'data': argument_edge}

                    if origin == 'steps':
                        execution_graph.add_edge(str(source), str(i))
                    else:
                        execution_graph.add_edge(origin, str(i))

            else: # PlaceholderPipelineDescriptionStep
                s = pipeline_description.steps[i].placeholder
                self.steptypes.append(StepType.PLACEHOLDER)
                self.primitives[i] = None
                self.subpipelines[i] = None
                for j in range(len(s.inputs)):
                    argument_edge = s.inputs[j].data
                    origin = argument_edge.split('.')[0]
                    source = argument_edge.split('.')[1]
                    self.primitives_arguments[i]['inputs'] = {'origin': origin, 'source': int(source), 'data': argument_edge}

                    if origin == 'steps':
                        execution_graph.add_edge(str(source), str(i))
                    else:
                        execution_graph.add_edge(origin, str(i))
            
        execution_order = list(nx.topological_sort(execution_graph))

        # Removing non-step inputs from the order
        execution_order = list(filter(lambda x: x.isdigit(), execution_order))
        self.execution_order = [int(x) for x in execution_order]

        # Creating set of steps to be call in produce
        self.produce_order = set()
        for i in range(len(pipeline_description.outputs)):
            output = pipeline_description.outputs[i]
            origin = output.data.split('.')[0]
            source = output.data.split('.')[1]
            self.outputs.append((origin, int(source), output.data, output.name))

            if origin != 'steps':
                continue
            else:
                current_step = int(source)
                self.produce_order.add(current_step)
                for i in range(0, len(execution_order)):
                    step_origin = self.primitives_arguments[current_step]['inputs']['origin']
                    step_source = self.primitives_arguments[current_step]['inputs']['source']
                    if step_origin != 'steps':
                        break
                    else:
                        self.produce_order.add(step_source)
                        current_step = step_source
        logging.info("Outputs = %s", self.outputs)

    def isDataFrameStep(self, n_step):
        if self.steptypes[n_step] is StepType.PRIMITIVE and \
        ('profiler' in self.primitives[n_step].metadata.query()['python_path']):
            return True
        return False

    def remove_one_hot_encoder(self):
        #For very high-dim datasets, replace one hot encoder primitive with "nothing".
        num = len(self.primitives)
        # Iterate through steps of the pipeline 
        for i in range(num):     
            p = self.primitives[i]
            python_path = p.metadata.query()['python_path'] 
            if 'one_hot_' in python_path:
                prim = d3m.index.get_primitive('d3m.primitives.data_transformation.do_nothing.DSBOX')
                self.primitives[i] = prim
                self.hyperparams[i] = None
                break

    def exclude(self, df):
        """
        Exclude columns which cannot/need not be processed.
        """
        self.exclude_columns = set()
        metadata = df.metadata
        rows = len(df)
        attributes = metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/Attribute")

        # Ignore all column that have only one value
        for att in attributes:
            attmeta = metadata.query((metadata_base.ALL_ELEMENTS, att))['semantic_types']
            length = len(attmeta)-1
            if 'https://metadata.datadrivendiscovery.org/types/UniqueKey' in attmeta:
                length = length - 1
                if length == 0:
                    self.exclude_columns.add(int(att))
                    continue
            col = int(att)

            # Remove constant attributes
            if len(df.iloc[:, col].unique()) <= 1:
                self.exclude_columns.add(col)

        cols = metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/FloatVector") # Found to be very expensive in ColumnParser!
        for col in cols:
            self.exclude_columns.add(col)

        if rows > 100000 and (len(attributes) - len(self.exclude_columns) > 50):
            random.seed(1979)
            remove = random.sample(attributes, len(attributes) - len(self.exclude_columns) - 50)
            self.exclude_columns.update(remove)
            self.remove_one_hot_encoder()
        if rows > 1000000 and (len(attributes) - len(self.exclude_columns) > 5):
            random.seed(1979)
            remove = random.sample(attributes, len(attributes) - len(self.exclude_columns) - 5)
            self.exclude_columns.update(remove)
            self.remove_one_hot_encoder() 
        if len(attributes) - len(self.exclude_columns) >= 10000:
            remove = random.sample(attributes, len(attributes) - len(self.exclude_columns) - 10000)
            self.exclude_columns.update(remove)
            self.remove_one_hot_encoder()

        targets = metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/TrueTarget")
        for t in targets:
            if t in self.exclude_columns:
                self.exclude_columns.remove(t)

    def produce_feature_importances(self, prefix):
        """
        Produce feature_importances
        """
        try:
            n_step = self.num_steps()-2
            model = self.pipeline[n_step]   
            step_name = "step."+str(n_step)+".produce_feature_importances"
            if prefix is not None:
                step_name = prefix + step_name
            values = model.produce_feature_importances().value
            return (step_name, values)   
        except:
            return (None, None)        

    def fit(self, **arguments):
        """
        Train all steps in the solution.

        Paramters
        ---------
        arguments
            Arguments required to train the solution
        """
        primitives_outputs = [None] * len(self.primitives) 
   
        if self.primitives_outputs is None: 
            for i in range(0, len(self.execution_order)):
                n_step = self.execution_order[i]
                primitives_outputs[n_step] = self.process_step(n_step, primitives_outputs, ActionType.FIT, arguments)
                if self.isDataFrameStep(n_step) == True:
                    self.exclude(primitives_outputs[n_step])
        else:
            last_step = self.get_last_step()
            primitives_outputs[last_step] = self.process_step(last_step, self.primitives_outputs, ActionType.FIT, arguments)
            i = 0
            for op in self.primitives_outputs:
                primitives_outputs[i] = op
                i = i + 1
            if last_step < len(self.execution_order) - 1:
                for i in range(last_step+1, len(self.execution_order)):
                     n_step = self.execution_order[i]
                     primitives_outputs[n_step] = self.process_step(n_step, primitives_outputs, ActionType.FIT, arguments)
            
        v = primitives_outputs[self.execution_order[len(self.execution_order)-1]]
        return v

    def _pipeline_step_fit(self, n_step: int, pipeline_id: str, primitive_arguments, arguments, action):
        """
        Execute a subpipeline step
        
        Paramters
        ---------
        n_step: int
            An integer of the actual step.
        pipeline_id: str
            
        primitive_arguments
            Arguments for the solution
        """
        solution = arguments['solution_dict'][pipeline_id]

        inputs = []
        inputs.append(primitive_arguments['inputs'])

        if action is ActionType.FIT: 
            return solution.fit(inputs=inputs, solution_dict=arguments['solution_dict'])
        elif action is ActionType.SCORE:
            return solution.score_solution(inputs=inputs, solution_dict=arguments['solution_dict'], primitive_dict=arguments['primitive_dict'],
                   posLabel=arguments['posLabel'], metric=arguments['metric'])
        else:
            return 

    def cross_validate_step(self, n_step, primitive: PrimitiveBaseMeta, primitive_arguments):
        # Build hyperparameters for the primitive
        custom_hyperparams = dict()
        hyperparams = self.hyperparams[n_step]
        if hyperparams is not None:
            for hyperparam, value in hyperparams.items():
                custom_hyperparams[hyperparam] = value

        training_arguments_primitive = self._primitive_arguments(primitive, 'set_training_data')
        training_arguments = {}

        for param, value in primitive_arguments.items():
            if param in training_arguments_primitive:
                training_arguments[param] = value

        # Instantiate the primitive (model)
        primitive_hyperparams = primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        prim_instance = primitive(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **custom_hyperparams))

        X, y = None, None
        y = training_arguments['outputs']
        X = training_arguments['inputs']

        task = "classification"
        if self.taskname != "classification":
            task = "regression"
        splits = get_num_splits(len(X), len(X.columns))
        split_indices = get_split_indices(X, y, splits, task)

        hyperparams = {}
        hyperparams['max_one_hot'] = 100
        prim = d3m.index.get_primitive('d3m.primitives.data_cleaning.robust_scaler.SKlearn')
        primitive_hyperparams = prim.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        scaler = prim(hyperparams=primitive_hyperparams.defaults())
        prim = d3m.index.get_primitive('d3m.primitives.data_transformation.to_numeric.DSBOX')
        primitive_hyperparams = prim.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        numeric = prim(hyperparams=primitive_hyperparams.defaults())

        train_predictions, test_predictions = [], []
        trainY, testY = [], []
        # Do the actual k-fold CV here
        for train_index, test_index in split_indices:
            X_train, X_test = X.iloc[train_index,:], X.iloc[test_index,:]
            y_train, y_test = y.iloc[train_index,:], y.iloc[test_index,:]

            X_test.reset_index(drop=True,inplace=True)
            X_train.reset_index(drop=True,inplace=True)
            prim_instance.set_training_data(inputs=X_train, outputs=y_train)
            prim_instance.fit()
            train = prim_instance.produce(inputs=X_train).value
            (cols, ordinals, add_floats) = helper_functions.get_cols_to_encode(train)
            hyperparams['use_columns'] = list(cols)
            prim = d3m.index.get_primitive('d3m.primitives.data_transformation.one_hot_encoder.DistilOneHotEncoder')
            primitive_hyperparams = prim.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
            ohe_prim = prim(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **hyperparams))
            ohe_prim.set_training_data(inputs=train)
            ohe_prim.fit()
            train = ohe_prim.produce(inputs=train).value
            train = numeric.produce(inputs=train).value
            scaler.set_training_data(inputs=train)
            scaler.fit()
            train = scaler.produce(inputs=train).value

            test = prim_instance.produce(inputs=X_test).value
            test = ohe_prim.produce(inputs=test).value
            test = numeric.produce(inputs=test).value
            test = scaler.produce(inputs=test).value
            train_predictions.append(train)
            test_predictions.append(test)
            trainY.append(y_train)
            testY.append(y_test)

        return [train_predictions, test_predictions, trainY, testY]
 
    def fit_step(self, n_step: int, primitive: PrimitiveBaseMeta, primitive_arguments):
        """
        Execute a primitive step

        Paramters
        ---------
        n_step: int
            An integer of the actual step.
        primitive: PrimitiveBaseMeta
            A primitive class
        primitive_arguments
            Arguments for set_training_data, fit, produce of the primitive for this step.
        """
        model = self.pipeline[n_step]
        produce_params_primitive = self._primitive_arguments(primitive, 'produce')
        produce_params = {}

        for param, value in primitive_arguments.items():
            if param in produce_params_primitive:
                produce_params[param] = value

        # If model is cached, just use it!
        python_path = primitive.metadata.query()['python_path']
        if model is not None:  # Use pre-learnt model
            return model.produce(**produce_params).value

        # Build hyperparameters for the primitive
        custom_hyperparams = dict() 
        hyperparams = self.hyperparams[n_step]
        if hyperparams is not None:
            for hyperparam, value in hyperparams.items():
                if hyperparam == 'primitive':
                    val = self.primitives[value]
                    primitive_hyperparams = val.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
                    primitive_model = val(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **self.hyperparams[value]))
                    custom_hyperparams[hyperparam] = primitive_model
                    continue
                custom_hyperparams[hyperparam] = value

        training_arguments_primitive = self._primitive_arguments(primitive, 'set_training_data')
        training_arguments = {}

        for param, value in primitive_arguments.items():
            if param in training_arguments_primitive:
                training_arguments[param] = value

        # Instantiate the primitive (model)
        primitive_hyperparams = primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        method_arguments = primitive.metadata.query()['primitive_code'].get('instance_methods', {}).get('__init__', {}).get('arguments', [])
        if 'volumes' in method_arguments:
            primitive_name = str(primitive)
            volumes = helper_functions.get_primitive_volumes(self.volumes_dir, primitive, primitive_name)
            model = primitive(volumes=volumes, hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **custom_hyperparams))
        else:
            model = primitive(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **custom_hyperparams))
    
        # Call set_training_data() on Primitive(model)
        model.set_training_data(**training_arguments)

        # Call fit() on Primitive(model)
        model.fit()

        # Reduce memory footprints!
        if 'splitter' in python_path or 'profiler' in python_path:
            model._training_inputs = None
        if 'corex' in python_path:
            model.training_data = None
        if 'tfidf' in python_path or 'robust_scaler' in python_path or 'imputer' in python_path or 'select_percentile' in python_path:
            model._inputs = None
            model._training_inputs = None
        if 'image_transfer' in python_path:
            self.pipeline[n_step] = None

        # Cache the model!
        self.pipeline[n_step] = model

        # Call produce() on Primitive(model)
        final_output = None

        # Some primitives have additional methods being called.
        if 'DistilRaggedDatasetLoader' in python_path or 'DistilAudioDatasetLoader' in python_path:
            final_output = {}
            final_output['produce'] = model.produce(**produce_params).value
            final_output['produce_collection'] = model.produce_collection(**produce_params).value
        elif 'DistilSingleGraphLoader' in python_path or \
             'DistilGraphLoader' in python_path or \
             'DistilEdgeListLoader' in python_path:
            final_output = {}
            final_output['produce'] = model.produce(**produce_params).value
            final_output['produce_target'] = model.produce_target(**produce_params).value
        else: # Most primitives have only "produce" method
            final_output = model.produce(**produce_params).value

        return final_output

    def _primitive_arguments(self, primitive, method: str) -> set:
        """
        Get the arguments of a primitive given a function.

        Parameters
        ---------
        primitive
            A primitive.
        method
            A method of the primitive.
        """
        return set(primitive.metadata.query()['primitive_code']['instance_methods'][method]['arguments'])

    def produce(self, **arguments):
        """
        Run produce on the solution.

        Paramters
        ---------
        arguments
            Arguments required to execute the solution
        """
        steps_outputs = [None] * len(self.primitives)

        # Loop through the pipeline
        for i in range(0, len(self.execution_order)):
            n_step = self.execution_order[i]
            produce_arguments = {}

            # Subpipeline
            if self.steptypes[n_step] is StepType.SUBPIPELINE:
                produce_arguments = {}
                for argument, value in self.primitives_arguments[n_step].items():
                    if value['origin'] == 'steps':
                        produce_arguments[argument] = steps_outputs[value['source']]
                    else:
                        produce_arguments[argument] = arguments['inputs'][value['source']]
                    if produce_arguments[argument] is None:
                        continue

            # Primitive
            if self.steptypes[n_step] is StepType.PRIMITIVE:
                primitive = self.primitives[n_step]
                python_path = primitive.metadata.query()['python_path']
                if self.pipeline[n_step] is None: #if 'resnet50' in python_path or self.pipeline[n_step] is None:
                    primitive_hyperparams = primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
                    custom_hyperparams = dict()

                    method_arguments = primitive.metadata.query()['primitive_code'].get('instance_methods', {}).get('__init__', {}).get('arguments', [])
                    if 'volumes' in method_arguments:
                        volumes = get_primitive_volumes(self.volumes_dir, primitive)
                        model = primitive(volumes=volumes, hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **custom_hyperparams))
                    else:
                        model = primitive(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **custom_hyperparams))

                    self.pipeline[n_step] = model
                produce_arguments_primitive = self._primitive_arguments(primitive, 'produce')
                for argument, value in self.primitives_arguments[n_step].items():
                    if argument in produce_arguments_primitive:
                        if value['origin'] == 'steps':
                            produce_arguments[argument] = steps_outputs[value['source']]
                        else:
                            produce_arguments[argument] = arguments['inputs'][value['source']]
                        if produce_arguments[argument] is None:
                            continue
            if self.steptypes[n_step] is StepType.PRIMITIVE: # Primitive
                if n_step in self.produce_order:
                    primitive = self.primitives[n_step]
                    v = self.pipeline[n_step].produce(**produce_arguments).value
                    steps_outputs[n_step] = v
                else:
                    steps_outputs[n_step] = None
            else: # Subpipeline
                solution_dict = arguments['solution_dict']
                solution = solution_dict[self.subpipelines[n_step]]
                inputs = []
                inputs.append(produce_arguments['inputs'])
                op = solution.produce(inputs=inputs, solution_dict=solution_dict)
                steps_outputs[n_step] = op[0]

        # Create output
        pipeline_output = []
        for output in self.outputs:
            if output[0] == 'steps':
                pipeline_output.append(steps_outputs[output[1]])
            else:
                pipeline_output.append(arguments[output[0][output[1]]])
        return pipeline_output

    def initialize_RPI_solution(self, taskname):
        """
        Initialize a solution from scratch consisting of predefined steps
        Leave last step for filling in primitive (for classification/regression problems)
        """
        python_paths = copy.deepcopy(solution_templates.task_paths[taskname])

        if 'denormalize' in python_paths[0] and self.ok_to_denormalize == False:
            python_paths.remove('d3m.primitives.data_transformation.denormalize.Common')

        self.taskname = taskname
        self.primitives_arguments = {}
        self.subpipelines = {}
        self.primitives = {}
        self.hyperparams = {}
        self.steptypes = []
        self.pipeline = []
        self.execution_order = None
    
        self.inputs = []
        self.inputs.append({"name": "dataset inputs"})

        # Constructing DAG to determine the execution order
        execution_graph = nx.DiGraph()

        # Setting steps relative to column_parser step
        col_parser_step = 3
        after_step = self.index_denormalize + 3
        if self.add_floats is not None and len(self.add_floats) > 0:
            python_paths.insert(after_step, 'd3m.primitives.data_transformation.add_semantic_types.Common')
            col_parser_step = 4

        num = len(python_paths)
        # Iterate through steps of the pipeline 
        for i in range(num):
            self.add_primitive(python_paths[i], i)

            # Set hyperparameters for specific primitives
            if python_paths[i] == 'd3m.primitives.data_transformation.add_semantic_types.Common':
                self.hyperparams[i] = {}
                exclude_atts = set() 
                if self.ordinal_atts is not None and len(self.ordinal_atts) > 0:
                    exclude_atts.update(list(self.ordinal_atts))
                if self.add_floats is not None and len(self.add_floats) > 0:
                    exclude_atts.update(list(self.add_floats))
                self.hyperparams[i]['columns'] = list(exclude_atts)
                self.hyperparams[i]['semantic_types'] = ['http://schema.org/Float']

            if python_paths[i] == 'd3m.primitives.data_cleaning.imputer.SKlearn':
                self.hyperparams[i] = {}
                self.hyperparams[i]['strategy'] = 'most_frequent'

            # Construct pipelines for different task types
            if i == 0: # denormalize
                data = 'inputs.0'
            elif i == col_parser_step + 1: # extract_columns_by_semantic_types (targets)
                data = 'steps.{}.produce'.format(self.index_denormalize + 2)
                self.hyperparams[i] = {}
                self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
            elif i == col_parser_step + 2: # extract_columns_by_semantic_types
                data = 'steps.{}.produce'.format(col_parser_step)
                self.hyperparams[i] = {}
                self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/Attribute','https://metadata.datadrivendiscovery.org/types/PrimaryKey']
                if self.privileged is not None and len(self.privileged) > 0:
                    self.hyperparams[i]['exclude_columns'] = self.privileged
            elif 'RPI' in python_paths[i]:
                data = 'steps.{}.produce'.format(col_parser_step + 1)
                origin = data.split('.')[0]
                source = data.split('.')[1]
                self.primitives_arguments[i]['outputs'] = {'origin': origin, 'source': int(source), 'data': data}
                execution_graph.add_edge(str(source), str(i))
                data = 'steps.' + str(i - 1) + str('.produce')
            else: # other steps
                data = 'steps.' + str(i - 1) + '.produce'

            origin = data.split('.')[0]
            source = data.split('.')[1]
            self.primitives_arguments[i]['inputs'] = {'origin': origin, 'source': int(source), 'data': data}

            if i == 0:
                execution_graph.add_edge(origin, str(i))
            else:
                execution_graph.add_edge(str(source), str(i))

        execution_order = list(nx.topological_sort(execution_graph))

        # Removing non-step inputs from the order
        execution_order = list(filter(lambda x: x.isdigit(), execution_order))
        self.execution_order = [int(x) for x in execution_order]

    def add_ssl_variant(self, variant):
        """
        Add SSL blackbox model as hyperparameter.
        Should support 'n_estimators' and predict_proba() API.
        """
        if variant == 'd3m.primitives.classification.gradient_boosting.SKlearn':
            from d3m.primitives.classification.gradient_boosting import SKlearn as blackboxParam
        elif variant == 'd3m.primitives.classification.extra_trees.SKlearn':
            from  d3m.primitives.classification.extra_trees import SKlearn as blackboxParam
        elif variant == 'd3m.primitives.classification.random_forest.SKlearn':
            from d3m.primitives.classification.random_forest import SKlearn as blackboxParam
        elif variant == 'd3m.primitives.classification.bagging.SKlearn':
            from d3m.primitives.classification.bagging import SKlearn as blackboxParam
        elif variant == 'd3m.primitives.classification.svc.SKlearn':
            from d3m.primitives.classification.svc import SKlearn as blackboxParam
        elif variant == 'd3m.primitives.classification.linear_svc.SKlearn':
            from d3m.primitives.classification.linear_svc import SKlearn as blackboxParam
 
        numSteps = self.num_steps()
        self.hyperparams[numSteps-2] = {}
        self.hyperparams[numSteps-2]['blackbox'] = blackboxParam

    def get_python_paths(self, taskname, columns, metric):
        """
        In addition to the pipeline template of steps, add extra feature processing steps (OHE/Scaler etc)
        """
        python_paths = copy.deepcopy(solution_templates.task_paths[taskname])

        if self.splitter == True:
            python_paths.insert(0, 'd3m.primitives.data_transformation.splitter.DSBOX')
            self.index_denormalize = 1

        if 'denormalize' in python_paths[self.index_denormalize] and self.ok_to_denormalize == False:
            python_paths.remove('d3m.primitives.data_transformation.denormalize.Common')

        if taskname in ['CLASSIFICATION','REGRESSION','COREXTEXT','IMAGE','IMAGE2','TIMESERIES','DISTILTEXT','TFIDFTEXT','SEMISUPERVISED']:
            if self.splitter == False and self.categorical_atts is not None and len(self.categorical_atts) > 0:
                #python_paths.append('d3m.primitives.data_transformation.one_hot_encoder.SKlearn')
                python_paths.append('d3m.primitives.data_transformation.one_hot_encoder.DistilOneHotEncoder')

            python_paths.append('d3m.primitives.data_transformation.to_numeric.DSBOX')
            if taskname not in ['IMAGE','IMAGE2','VIDEO','TFIDFTEXT']:
                # Data frame has too many dimensions (in thousands)! This step is extremely slow! 
                python_paths.append('d3m.primitives.data_cleaning.robust_scaler.SKlearn')

            # ColumnParser.Common is expensive!
            if columns > 2000:  
                python_paths[self.index_denormalize+4] = 'd3m.primitives.data_transformation.column_parser.DistilColumnParser'
                python_paths.append('d3m.primitives.feature_selection.select_percentile.SKlearn')
       
            # Add float semantic type to Unknown attributes 
            after_target_step = self.index_denormalize + 4
            if (self.add_floats is not None and len(self.add_floats) > 0) or (self.ordinal_atts is not None and len(self.ordinal_atts) > 0):
                python_paths.insert(after_target_step, 'd3m.primitives.data_transformation.add_semantic_types.Common')

            if metric is not None and 'ROC_AUC' in metric:
                python_paths.append('d3m.primitives.data_transformation.add_semantic_types.Common')
                python_paths.append('d3m.primitives.data_transformation.horizontal_concat.DataFrameCommon')
                python_paths.append('d3m.primitives.operator.compute_unique_values.Common')
                python_paths.append('d3m.primitives.data_transformation.extract_columns_by_semantic_types.Common')
                python_paths.append('d3m.primitives.data_transformation.extract_columns_by_semantic_types.Common')
            
        return python_paths

    def initialize_solution(self, taskname, columns=10, metric=None):
        """
        Initialize a solution from scratch consisting of predefined steps
        Pipelines constructed from templates specified in solution_templates.py
        Leave last step for filling in primitive (for classification/regression problems)
        """
        self.taskname = taskname
        python_paths = self.get_python_paths(taskname, columns, metric)

        self.primitives_arguments = {}
        self.subpipelines = {}
        self.primitives = {}
        self.hyperparams = {}
        self.steptypes = []
        self.pipeline = []
        self.execution_order = None

        self.inputs = []
        self.inputs.append({"name": "dataset inputs"})

        # Constructing DAG to determine the execution order
        execution_graph = nx.DiGraph()

        num = len(python_paths)

        # Iterate through steps of the pipeline 
        for i in range(num):
            self.add_primitive(python_paths[i], i)

            # Set hyperparameters for specific primitives
            self.hyperparams[i] = default_hyperparams.set_default_hyperparameters(python_paths[i], taskname)
            if 'percentile' in python_paths[i]:
                self.hyperparams[i] = {}
                percent =  int(math.floor(100*100/columns))
                if percent < 1:
                    percent = 1
                self.hyperparams[i]['percentile'] = percent
                if 'REGRESSION' in taskname:
                    self.hyperparams[i]['score_func'] = 'f_regression'

            if python_paths[i] == 'd3m.primitives.data_transformation.add_semantic_types.Common':
                self.hyperparams[i] = {}
                exclude_atts = set()
                if self.ordinal_atts is not None and len(self.ordinal_atts) > 0:
                    exclude_atts.update(list(self.ordinal_atts))
                if self.add_floats is not None and len(self.add_floats) > 0:
                    exclude_atts.update(list(self.add_floats))
                self.hyperparams[i]['columns'] = list(exclude_atts)
                self.hyperparams[i]['semantic_types'] = ['http://schema.org/Float', 'https://metadata.datadrivendiscovery.org/types/Attribute']
                
            if self.privileged is not None and len(self.privileged) > 0 and \
                python_paths[i] == 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common':
                self.hyperparams[i] = {}
                self.hyperparams[i]['exclude_columns'] = self.privileged
           
            if 'concat' in python_paths[i]:
                data = 'steps.{}.produce'.format(i-1)
                self.primitives_arguments[i]['left'] = get_argument(data) 
                data = 'steps.{}.produce'.format(self.index_denormalize + 3)
                self.primitives_arguments[i]['right'] = get_argument(data)    
                execution_graph.add_edge(str(i-1), str(i))
                execution_graph.add_edge(str(self.index_denormalize + 3), str(i))

            # Construct pipelines for different task types
            if taskname in ['CLASSIFICATION','REGRESSION','COREXTEXT','IMAGE','IMAGE2','VIDEO','TIMESERIES','DISTILTEXT','TFIDFTEXT','CONDITIONER','FORE_REGRESSION']:
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == self.index_denormalize + 3 or (metric is not None and 'ROC_AUC' in metric and i == num-2): # extract_columns_by_semantic_types (targets)
                    data = 'steps.{}.produce'.format(i-1)
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == self.index_denormalize + 4 or (metric is not None and 'ROC_AUC' in metric and i == num-1): # column_parser/time series/add_semantic_types
                    data = 'steps.{}.produce'.format(i-2)
                elif 'DistilTextEncoder' in python_paths[i] or 'select_percentile' in python_paths[i]: # DistilTextEncoder / select_percentile: These need inputs AND outputs!
                    data = 'steps.{}.produce'.format(self.index_denormalize + 3)
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    source = self.primitives_arguments[i]['outputs']['source']
                    execution_graph.add_edge(str(source), str(i))
                    data = 'steps.' + str(i - 1) + str('.produce')
                else: # other steps
                    data = 'steps.' + str(i - 1) + '.produce'
            elif taskname == 'TIMESERIES3':
                if i == 0: 
                    data = 'inputs.0'
                else: # other steps
                    data = 'steps.0.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    data = 'steps.' + str(i - 1) + '.produce'

            # Graph problems!
            elif taskname in ['COMMUNITYDETECTION','LINKPREDICTION','VERTEXCLASSIFICATION','VERTEXCLASSIFICATION3','GRAPHMATCHING']:
                if i == 0: 
                    data = 'inputs.0'
                else:
                    data = 'steps.0.produce_target'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    data = 'steps.0.produce'

            # Another graph pipeline
            elif taskname == 'GRAPHMATCHING3':
                if i == 0 or i == 1 or i == 2: # dataset_to_dataframe
                    data = 'inputs.0'
                    if i > 0:
                        self.hyperparams[i] = {}
                        self.hyperparams[i]['dataframe_resource'] = str(i)
                else: #euclidean_nomination.JHU
                    data = 'steps.1.produce'
                    self.primitives_arguments[i]['inputs_1'] = get_argument(data)
                    data = 'steps.2.produce'
                    self.primitives_arguments[i]['inputs_2'] = get_argument(data)
                    data = 'steps.0.produce'
                    self.primitives_arguments[i]['reference'] = get_argument(data)
                    execution_graph.add_edge(str(0), str(i))
                    execution_graph.add_edge(str(1), str(i))
                    execution_graph.add_edge(str(2), str(i))

            # ISI_PIPELINE
            elif taskname == 'ISI_PIPELINE':
                if i == 0:
                    data = 'inputs.0'
                elif i == 2: # extract_columns_by_semantic_types (targets)
                    data = 'steps.{}.produce'.format(i-1)
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 3: # gcn
                    data = 'steps.2.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    source = self.primitives_arguments[i]['outputs']['source']
                    execution_graph.add_edge(str(source), str(i))
                    data = 'steps.0.produce'

            # Audio
            elif taskname == 'AUDIO':
                if i == 0: # AudioDatasetLoader
                    data = 'inputs.0'
                elif i == self.index_denormalize + 1: # column_parser
                    data = 'steps.{}.produce'.format(self.index_denormalize)
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['parse_semantic_types'] = ["http://schema.org/Boolean", \
                                                                   "http://schema.org/Integer", \
                                                                   "http://schema.org/Float"]
                elif i == self.index_denormalize + 2: # extract_columns_by_semantic_types (targets)
                    data = 'steps.{}.produce'.format(self.index_denormalize + 1)
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == self.index_denormalize + 3: # DistilAudioTransfer
                    data = 'steps.{}.produce_collection'.format(self.index_denormalize)
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # Remote-sensing
            elif taskname == 'REMOTESENSING':
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == 3: # DistilColumnParser
                    data = 'steps.2.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['parse_semantic_types'] = ["http://schema.org/Integer", \
                                                                   "http://schema.org/Float", \
                                                                   "https://metadata.datadrivendiscovery.org/types/FloatVector"]
                elif i == 4: # extract_columns_by_semantic_types 
                    data = 'steps.3.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ["http://schema.org/ImageObject"]
                elif i == 5: # extract_columns_by_semantic_types (targets)
                    data = 'steps.3.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 6: # remote_sensing
                    data = 'steps.4.produce'
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # Semi-supervised
            elif taskname == 'SEMISUPERVISED' or taskname == 'SEMISUPERVISED_HDB':
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == 3: # extract_columns_by_semantic_types (targets)
                    data = 'steps.2.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 4: # column_parser
                    data = 'steps.2.produce'
                elif 'extract_columns' in python_paths[i]:
                    data = 'steps.' + str(i-1) + '.produce'
                    if self.hyperparams[i] is None:
                        self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/Attribute', \
                                                             'https://metadata.datadrivendiscovery.org/types/PrimaryKey']
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # ARIMA
            elif taskname == 'ARIMA_FORECASTING':
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == 3: # extract_columns_by_semantic_types (targets)
                    data = 'steps.2.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 4: # column_parser
                    data = 'steps.2.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['parse_semantic_types'] = ["http://schema.org/Boolean", \
                                                                   "http://schema.org/Integer", \
                                                                   "http://schema.org/Float"]
                elif i == 5: # extract_columns
                    data = 'steps.4.produce'
                    if self.hyperparams[i] is None:
                        self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/Attribute', \
                                                             'https://metadata.datadrivendiscovery.org/types/PrimaryKey', \
                                                             'https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 6: # arima
                    data = 'steps.3.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    source = self.primitives_arguments[i]['outputs']['source']
                    data = 'steps.' + str(i-1) + '.produce'
                    execution_graph.add_edge(str(source), str(i))
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # Other forecasting pipelines!
            elif taskname in ['ESRNN_FORECASTING','NBEATS_FORECASTING','DISTIL_NBEATS']:
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == 4: # extract_columns_by_semantic_types (targets)
                    data = 'steps.3.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 5: # extract_columns
                    data = 'steps.3.produce'
                elif i == 3: # column_parser
                    data = 'steps.' + str(2) + '.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['parse_semantic_types'] = ["http://schema.org/Boolean", \
                                                                   "http://schema.org/Integer", \
                                                                   "http://schema.org/Float", \
                                                                   "https://metadata.datadrivendiscovery.org/types/FloatVector", \
                                                                   "http://schema.org/DateTime"]
                elif 'esrnn' in python_paths[i] or 'nbeats' in python_paths[i] or 'NBEATS' in python_paths[i]: # esrnn / nbeats
                    data = 'steps.4.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    source = self.primitives_arguments[i]['outputs']['source']
                    data = 'steps.' + str(i-1) + '.produce'
                    execution_graph.add_edge(str(source), str(i))
                elif 'construct_predictions' in python_paths[i]: #i == 9:
                    data = 'steps.1.produce'
                    self.primitives_arguments[i]['reference'] = get_argument(data)
                    data = 'steps.' + str(i-1) + '.produce'
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # Other forecasting pipelines!
            elif taskname == 'VAR_FORECASTING' or taskname == 'LSTM_FORECASTING':
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == 4: # extract_columns_by_semantic_types (targets)
                    data = 'steps.3.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 3: # column_parser
                    data = 'steps.' + str(2) + '.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['parse_semantic_types'] = ["http://schema.org/Boolean", \
                                                                   "http://schema.org/Integer", \
                                                                   "http://schema.org/Float", \
                                                                   "https://metadata.datadrivendiscovery.org/types/FloatVector", \
                                                                   "http://schema.org/DateTime"]
                elif i == 5: # extract_columns
                    data = 'steps.3.produce'
                    if self.hyperparams[i] is None:
                        self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/Attribute', \
                                                             'https://metadata.datadrivendiscovery.org/types/PrimaryKey']
                elif 'AR' in python_paths[i]: # VAR
                    data = 'steps.' + str(4) + '.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    source = self.primitives_arguments[i]['outputs']['source']
                    execution_graph.add_edge(str(source), str(i))
                    data = 'steps.' + str(i-1) + '.produce'
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # Object detection
            elif taskname == 'OBJECTDETECTION':
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == 2: # extract_columns_by_semantic_types
                    data = 'steps.1.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/PrimaryMultiKey', \
                                                             'https://metadata.datadrivendiscovery.org/types/FileName']
                elif i == 3: # extract_columns_by_semantic_types (targets)
                    data = 'steps.1.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 4: # yolo
                    data = 'steps.3.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    data = 'steps.2.produce'
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # Another object detection pipeline
            elif taskname == 'OBJECTDETECTION2':
                if i == 0: # denormalize
                    data = 'inputs.0'
                elif i == 2: # retina_net
                    data = 'steps.1.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce' 

            # Collaborative filtering
            elif taskname == 'COLLABORATIVEFILTERING':
                if i == 0: # dataset_to_dataframe
                    data = 'inputs.0'
                elif i == num-1: # construct_predictions
                    data = 'steps.1.produce'
                    self.primitives_arguments[i]['reference'] = get_argument(data)
                    data = 'steps.' + str(i-1) + '.produce'
                elif i == 4: # extract_columns_by_semantic_types (targets)
                    data = 'steps.3.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/TrueTarget']
                elif i == 5: # extract_columns_by_semantic_types
                    data = 'steps.3.produce'
                elif i == 3:
                    data = 'steps.' + str(i-1) + '.produce'
                    self.hyperparams[i] = {}
                    self.hyperparams[i]['parse_semantic_types'] = ["http://schema.org/Integer", \
                                                                   "http://schema.org/Float"]
                elif i == 6: # collaborative_filtering
                    data = 'steps.4.produce'
                    self.primitives_arguments[i]['outputs'] = get_argument(data)
                    source = self.primitives_arguments[i]['outputs']['source']
                    execution_graph.add_edge(str(source), str(i))
                    data = 'steps.' + str(i-1) + '.produce'
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'

            # Clustering
            elif taskname == 'CLUSTERING':
                if i == 0: # dataset_to_dataframe
                    data = 'inputs.0'
                elif i == num-1: # construct_predictions
                    data = 'steps.0.produce'
                    self.primitives_arguments[i]['reference'] = get_argument(data)
                    data = 'steps.' + str(i-1) + '.produce'
                else: # other steps
                    data = 'steps.' + str(i-1) + '.produce'
            else:
                if i == 0:
                    data = 'inputs.0'
                else:
                    data = 'steps.' + str(i-1) + '.produce'

            self.primitives_arguments[i]['inputs'] = get_argument(data)
            source = self.primitives_arguments[i]['inputs']['source']
            origin = self.primitives_arguments[i]['inputs']['origin']

            if i == 0:
                execution_graph.add_edge(origin, str(i))
            else:
                execution_graph.add_edge(str(source), str(i))

            argument_metadata = self.primitives[i].metadata.query()['primitive_code'].get('arguments', {})
            if argument_metadata.get("inputs", None) is None:
                self.primitives_arguments[i].pop("inputs", None)
        execution_order = list(nx.topological_sort(execution_graph))
        # Removing non-step inputs from the order
        execution_order = list(filter(lambda x: x.isdigit(), execution_order))
        self.execution_order = [int(x) for x in execution_order]

    def add_primitive(self, python_path, i):
        """
        Helper function to add a primitive in the pipeline with basic initialization only.
        """
        self.primitives_arguments[i] = {}
        self.hyperparams[i] = None
        self.pipeline.append(None)
        prim = d3m.index.get_primitive(python_path)
        self.primitives[i] = prim
        self.steptypes.append(StepType.PRIMITIVE)
        self.subpipelines[i] = None

    def add_subpipeline(self, pipeline):
        """
        Helper function to add a subpipeline(replace placeholder step) in the pipeline with basic initialization only.
        """
        i = len(self.primitives_arguments)-1

        self.primitives_arguments[i] = {}
        self.hyperparams[i] = None
        self.pipeline.append(None)
        self.primitives[i] = None
        self.steptypes[i] = StepType.SUBPIPELINE
        self.subpipelines[i] = pipeline.id
      
        data = 'steps.' + str(i-1) + '.produce'
        origin = data.split('.')[0]
        source = data.split('.')[1] 
        self.primitives_arguments[i]['inputs'] = {'origin': origin, 'source': int(source), 'data': data}

        data = 'steps.' + str(i) + '.' + pipeline.outputs[0][2]
        origin = data.split('.')[0]
        source = data.split('.')[1]
        self.outputs = []
        self.outputs.append((origin, int(source), data, "output predictions"))

    def add_step(self, python_path, inputstep=-1, outputstep=2, dataframestep=1, construct_primitive='d3m.primitives.data_transformation.construct_predictions.Common'):
        """
        Add new primitive (classifier/regressor)
        This is currently for classification/regression pipelines
        python_path: Primitive
        inputstep: Step index of inputs for the primitive
        outputstep: Step index of outputs for the primitive
        dataframestep: Step index of data frame for the construct_predictions/construct confidence primitive's reference (to get d3mIndex)
        """
        i = len(self.primitives_arguments)
        self.add_primitive(python_path, i)

        if inputstep == -1:
            inputstep = i-1
        # Connect to inputs
        data = 'steps.' + str(inputstep) + str('.produce')
        origin = data.split('.')[0]
        source = data.split('.')[1]
        self.primitives_arguments[i]['inputs'] = {'origin': origin, 'source': int(source), 'data': data}
        
        # Connect to outputs
        data = 'steps.' + str(outputstep) + str('.produce') # extract_columns_by_semantic_types (targets)
        origin = data.split('.')[0]
        source = data.split('.')[1]
        self.primitives_arguments[i]['outputs'] = {'origin': origin, 'source': int(source), 'data': data}

        if 'learner.random_forest' in python_path:
            self.hyperparams[i] = {}
            self.hyperparams[i]['compute_confidences'] = True
            self.hyperparams[i]['n_estimators'] = 100

        if 'SKlearn' in python_path:
            self.hyperparams[i] = {}
            hyperparam_spec = self.primitives[i].metadata.query()['primitive_code']['hyperparams']
            if 'n_estimators' in hyperparam_spec:
                self.hyperparams[i]['n_estimators'] = 100
            if 'n_neighbors' in hyperparam_spec:
                self.hyperparams[i]['n_neighbors'] = 10

        self.execution_order.append(i)

        i = i + 1
        self.add_primitive(construct_primitive, i)

        data = 'steps.' + str(i-1) + str('.produce')
        if 'construct_confidence' in construct_primitive: # For ROC_AUC metric! 
            data = 'steps.' + str(i-4) + str('.produce')  # Should point to d3m.primitives.operator.compute_unique_values.Common

        origin = data.split('.')[0]
        source = data.split('.')[1]
        self.primitives_arguments[i]['inputs'] = {'origin': origin, 'source': int(source), 'data': data}
        
        data = 'steps.' + str(dataframestep) + str('.produce')
        origin = data.split('.')[0]
        source = data.split('.')[1]
        self.primitives_arguments[i]['reference'] = {'origin': origin, 'source': int(source), 'data': data}

        if 'construct_confidence' in construct_primitive:
            self.hyperparams[i] = {}
            self.hyperparams[i]['primitive_learner'] = i-1

        self.execution_order.append(i)
        self.add_outputs()

    def splitter_present(self):
        python_path = self.primitives[0].metadata.query()['python_path']
        if 'splitter' in python_path:
            return True
        return False

    def complete_solution(self, optimal_params):
        """
        This is currently for classification/regression pipelines with AutoRPI
        Sets optimal hyperparamters found for AutoRPI primitive and the model primitive.
        """
        i = self.get_last_step()
        self.hyperparams[i] = {}
        self.hyperparams[i]['method'] = optimal_params[1]
        self.hyperparams[i]['nbins'] = optimal_params[0]

        i = i + 2
        python_path = self.primitives[i].metadata.query()['python_path']
        if 'linear_disc' not in python_path:
            self.hyperparams[i] = {}
            self.hyperparams[i]['n_estimators'] = optimal_params[2]
        if 'gradient_boosting' in python_path:
            self.hyperparams[i]['learning_rate'] = 10/optimal_params[2]

    def add_RPI_step(self, RPI_step, python_path, output_step):
        """
        Complete pipeline with RPIprimitive and ML model
        This is currently for classification/regression pipelines with d3m.primitives.feature_selection.joint_mutual_information.AutoRPI
        and d3m.primitives.feature_selection.simultaneous_markov_blanket.AutoRPI 
        RPI_step: RPI primitive (Either d3m.primitives.feature_selection.joint_mutual_information.AutoRPI or d3m.primitives.feature_selection.simultaneous_markov_blanket.AutoRPI)
        python_path: ML model being added (Classifier/regressor)
        output_step: Step index of targets in the pipeline
        """
        n_steps = len(self.primitives_arguments) + 1
        i = n_steps-1

        # Add RPI step
        self.add_primitive(RPI_step, i)

        data = 'steps.' + str(i-1) + str('.produce')
        self.primitives_arguments[i]['inputs'] = get_argument(data)

        data = 'steps.' + str(output_step) + str('.produce') # extract_columns_by_semantic_types (targets)
        self.primitives_arguments[i]['outputs'] = get_argument(data)

        self.execution_order.append(i)

        # Add imputer
        i = i + 1
        self.add_primitive('d3m.primitives.data_cleaning.imputer.SKlearn', i)
        data = 'steps.' + str(i-1) + str('.produce')
        self.primitives_arguments[i]['inputs'] = get_argument(data)
        self.hyperparams[i] = {}
        self.hyperparams[i]['strategy'] = 'most_frequent'
        self.execution_order.append(i)

        # Add ML model
        i = i + 1
        self.add_primitive(python_path, i)
        data = 'steps.' + str(i-1) + str('.produce')
        self.primitives_arguments[i]['inputs'] = get_argument(data)
        data = 'steps.' + str(output_step) + str('.produce') # extract_columns_by_semantic_types (targets)
        self.primitives_arguments[i]['outputs'] = get_argument(data)
        self.execution_order.append(i)

        # Add construct_predictions
        i = i + 1
        self.add_primitive('d3m.primitives.data_transformation.construct_predictions.Common', i)
        data = 'steps.' + str(i-1) + str('.produce')
        self.primitives_arguments[i]['inputs'] = get_argument(data) 
        data = 'steps.' + str(1) + str('.produce')
        self.primitives_arguments[i]['reference'] = get_argument(data) 
        self.execution_order.append(i)

        self.add_outputs()

    def add_outputs(self): 
        """
        Add outputs as last step for pipeline
        Also compute produce order.
        """
        n_steps = len(self.execution_order)
        
        self.outputs = []

        data = 'steps.' + str(self.execution_order[n_steps-1]) + '.produce'

        origin = data.split('.')[0]
        source = data.split('.')[1]
        self.outputs.append((origin, int(source), data, "output predictions"))

        # Creating set of steps to be call in produce
        self.produce_order = set()

        current_step = int(source)
        self.produce_order.add(current_step)

        # Iterate through the pipeline execution order
        for i in range(0, len(self.execution_order)):
            name = list(self.primitives_arguments[current_step].keys())[0]
            step_origin = self.primitives_arguments[current_step][name]['origin']
            step_source = self.primitives_arguments[current_step][name]['source']
            if step_origin != 'steps':
                break
            else:
                self.produce_order.add(step_source)
                current_step = step_source

        data = 'steps.' + str(self.execution_order[n_steps-1]) + '.produce'
        source = data.split('.')[1]
        current_step = int(source)

        # Iterate once more. Steps to 'reference' argument should not get missed!
        for i in range(0, len(self.execution_order)):
            if 'reference' in self.primitives_arguments[current_step]:
                step_source = self.primitives_arguments[current_step]['reference']['source']
                if step_source in self.produce_order:
                    break
                else:
                    self.produce_order.add(step_source)
                    current_step = step_source

                    for j in range(0, len(self.execution_order)):
                        name = list(self.primitives_arguments[current_step].keys())[0]
                        step_origin = self.primitives_arguments[current_step][name]['origin']
                        step_source = self.primitives_arguments[current_step][name]['source']
                        if step_origin != 'steps':
                            break
                        else:
                            self.produce_order.add(step_source)
                            current_step = step_source

    def process_step(self, n_step, primitives_outputs, action, arguments):
        """
        Process each step of a pipeline
        This could be used while scoring or fitting a pipeline
        """
        # Subpipeline step
        if self.steptypes[n_step] is StepType.SUBPIPELINE:
            primitive_arguments = {}
            for argument, value in self.primitives_arguments[n_step].items():
                if value['origin'] == 'steps':
                    primitive_arguments[argument] = primitives_outputs[value['source']]
                else:
                    primitive_arguments[argument] = arguments['inputs'][value['source']]
            return self._pipeline_step_fit(n_step, self.subpipelines[n_step], primitive_arguments, arguments, action)

        # Primitive step
        if self.steptypes[n_step] is StepType.PRIMITIVE:
            primitive_arguments = {}
            python_path = self.primitives[n_step].metadata.query()['python_path']
            for argument, value in self.primitives_arguments[n_step].items():
                if value['origin'] == 'steps':
                    # Primitives with multiple produce methods used in pipeline
                    if 'DistilLinkPrediction' in python_path or \
                       'DistilSeededGraphMatcher' in python_path or \
                       'DistilCommunityDetection' in python_path or \
                       'DistilVertexNomination' in python_path or \
                       'DistilEdgeListLoader' in python_path or \
                       'DistilAudioTransfer' in python_path or \
                       'DistilSingleGraphLoader' in python_path or \
                       'DistilGraphLoader' in python_path or \
                       (self.taskname == 'AUDIO' and 'column_parser' in python_path):
                        method = value['data'].split('.')[2]
                        primitive_arguments[argument] = primitives_outputs[value['source']][method]
                    else: # Single output
                        primitive_arguments[argument] = primitives_outputs[value['source']]
                else:
                    primitive_arguments[argument] = arguments['inputs'][value['source']]

            # Score the pipeline at model stage (Run cross-validation)
            if action is ActionType.SCORE and self.is_last_step(n_step) == True: 
                primitive = self.primitives[n_step]
                primitive_desc = arguments['primitive_dict'][primitive]
                mode = None
                if 'mode' in arguments:
                    mode = arguments['mode']
                return self.score_step(primitive, primitive_arguments, arguments['metric'], arguments['posLabel'], primitive_desc, \
                                       self.hyperparams[n_step], n_step, mode)
            else: # Fit the pipeline step
                primitive = self.primitives[n_step]
                if 'DistilTextEncoder' in python_path and self.pipeline[n_step] is None:
                    self.cross_outputs = self.cross_validate_step(n_step, primitive, primitive_arguments)
                v = self.fit_step(n_step, primitive, primitive_arguments)
                return v

        # Placeholder step
        if self.steptypes[n_step] is StepType.PLACEHOLDER:
            return primitives_outputs[n_step-1]
 
    def is_last_step(self, n):
        last_step = self.get_last_step()
        if n == self.execution_order[last_step]:
            return True
        return False

    def get_last_step(self):
        """
        Return index of step which needs to be optimized/cross-validated.
        Typically this is the step for a classifier/regressor which can be optimized/cross-validated.
        This is the last-1 step of the pipeline, since it is followed by construct_predictions primitive to construct predictions output from d3mIndex and primitive output.
        """
        n_steps = len(self.execution_order)
        last_step = self.execution_order[n_steps-1]

        if self.steptypes[last_step] is StepType.SUBPIPELINE:
            return n_steps-1

        if self.primitives[last_step] is None:
            return n_steps-1

        if self.taskname == 'PIPELINE_RPI':
            index = 0
            for p in self.primitives:
                python_path = None
                if self.primitives[index] is not None:
                    python_path = self.primitives[index].metadata.query()['python_path'] 
                    if 'RPI' in python_path:
                        return index
                index = index + 1
 
        last_step = len(self.primitives)-1
        if 'construct_' in self.primitives[last_step].metadata.query()['python_path']:
            return last_step-1
        else:
            return last_step

    def score_solution(self, **arguments):
        """
        Score a solution 
        """
        score = 0.0
        primitives_outputs = [None] * len(self.primitives)

        last_step = self.get_last_step()

        if self.primitives_outputs is None: # Run entire pipeline
            for i in range(0, last_step+1):
                n_step = self.execution_order[i]
                primitives_outputs[n_step] = self.process_step(n_step, primitives_outputs, ActionType.SCORE, arguments)
                if self.isDataFrameStep(n_step) == True:
                    self.exclude(primitives_outputs[n_step])
        else: # Cached outputs. Score the ML model only
            primitives_outputs[last_step] = self.process_step(last_step, self.primitives_outputs, ActionType.SCORE, arguments)

        (score, optimal_params) = primitives_outputs[self.execution_order[last_step]]
        if self.steptypes[self.execution_order[last_step]] != StepType.SUBPIPELINE:
            self.hyperparams[last_step] = optimal_params
        
        return (score, optimal_params)

    def set_distil_text_hyperparam(self):
        """
        For text regression problems, set appropriate metric for DistilTextEncoder
        """
        for i in range(len(self.primitives)):
            p = self.primitives[i]
            path = p.metadata.query()['python_path'] 
            if 'DistilTextEncoder' in path:
                self.hyperparams[i] = {}
                self.hyperparams[i]['metric'] = 'meanSquaredError'
                break

    def set_classifier_hyperparam(self):
        for i in range(len(self.primitives)):
            p = self.primitives[i]
            path = p.metadata.query()['python_path']
            if 'random_forest' in path or 'extra_trees' in path or 'gradient_boosting' in path:
                self.hyperparams[i] = {}
                self.hyperparams[i]['n_estimators'] = 5
                break

    def set_hyperparams(self, solution_dict, hp):
        """
        Set hyperparameters for the primtiive at the "last" step.
        """
        n_step = self.get_last_step()

        if self.primitives[self.execution_order[n_step]] is not None: 
            if 'RPI' in self.primitives[self.execution_order[n_step]].metadata.query()['python_path']:
                self.complete_solution(hp)
            else:
                self.hyperparams[self.execution_order[n_step]] = hp
        else:
            solution_id = self.subpipelines[self.execution_order[n_step]]
            solution_dict[solution_id].set_hyperparams(solution_dict, hp)

        self.pipeline_description = None #Recreate it again

    def run_basic_solution(self, **arguments):
        """
        Runs common parts of a pipeline before adding last step of classifier/regressor
        This saves on data processing, featurizing steps being repeated across multiple pipelines.
        This is probably the most IMPORTANT step for multimedia datasets(audio/video/images), text datasets, or large datasets.
        Runs single-pass on expensive data loading, processing steps before forking multiple processes for validating ML models. 
        Forked processes get cached outputs and work on cross-validating the ML models only.
        """
        self.primitives_outputs = [None] * len(self.execution_order)

        output_step = arguments['output_step']
        if 'dataframe_step' in arguments:
            dataframe_step = arguments['dataframe_step']
        else:
            dataframe_step = self.index_denormalize + 1

        if 'input_step' in arguments:
            input_step = arguments['input_step']
        else:
            input_step = len(self.execution_order)-1

        # Execute the initial steps of a pipeline.
        # This executes all the common data processing steps of classifier/regressor pipelines, but only once for all pipelines.
        # Inputs and outputs for the next step (classifier/regressor) are stored.
        for i in range(0, len(self.execution_order)):
            n_step = self.execution_order[i]
            python_path = self.primitives[n_step].metadata.query()['python_path']
       
            if 'one_hot_encoder' in python_path:
                (cols, ordinals, add_floats) = helper_functions.get_cols_to_encode(self.primitives_outputs[n_step-1])
                self.hyperparams[n_step] = {}
                self.hyperparams[n_step]['use_columns'] = list(cols)
                self.hyperparams[n_step]['max_one_hot'] = 100
                print("Cats = ", cols)
                #print(self.primitives_outputs[n_step-1].iloc[:,cols])

            if python_path == 'd3m.primitives.data_transformation.column_parser.Common':
                if self.hyperparams[n_step] is None:
                    self.hyperparams[n_step] = {}
                exclude_atts = set() #
                if self.ordinal_atts is not None and len(self.ordinal_atts) > 0:
                    exclude_atts.update(list(self.ordinal_atts))
                if self.exclude_columns is not None and len(self.exclude_columns) > 0:
                    exclude_atts.update(list(self.exclude_columns))
                if len(exclude_atts) > 0:
                    self.hyperparams[n_step]['exclude_columns'] = list(exclude_atts)

            if self.exclude_columns is not None and len(self.exclude_columns) > 0:
                if python_path == 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common':
                    if self.hyperparams[n_step] is None:
                        self.hyperparams[n_step] = {}
                    exclude_atts = set()
                    if 'exclude_columns' in self.hyperparams[n_step]:
                        exclude_atts.update(self.hyperparams[n_step]['exclude_columns'])
                    exclude_atts.update(list(self.exclude_columns))
                    self.hyperparams[n_step]['exclude_columns'] = list(exclude_atts)
                    self.exclude_columns = None

            if 'add_semantic_types' in python_path:
                from_step = int(self.primitives_arguments[n_step]['inputs']['source'])
                from_python_path = self.primitives[from_step].metadata.query()['python_path']
                if 'profiler' not in from_python_path:
                    self.hyperparams[n_step]['columns'] = list(range(len(self.primitives_outputs[from_step].columns)))
        
            # Running the primitive at n_step and cache the output
            logging.critical("Running %s", python_path)
            start = timer()
            self.primitives_outputs[n_step] = self.process_step(n_step, self.primitives_outputs, ActionType.FIT, arguments)
            end = timer()
            logging.critical("Time taken : %s sec", end - start)
            #if n_step > 1:
            #    logging.critical("%s", self.primitives_outputs[n_step].iloc[0:5,:])
            if self.isDataFrameStep(n_step) == True:
                self.exclude(self.primitives_outputs[n_step])

        # Remove other intermediate step outputs, they are not needed anymore.
        # Only inputs, outputs for training model are retained.
        # Also dataframe for constructing predictions.
        for i in range(0, len(self.execution_order)):
            if i == dataframe_step or i == output_step or i == input_step:
                continue
            self.primitives_outputs[i] = [None]

        # Randomly sample 100K points for scoring solutions
        if len(self.primitives_outputs[output_step]) > 1000000:
            np.random.seed(1979)
            indices = np.random.choice(len(self.primitives_outputs[output_step]),100000, replace=False)
            self.primitives_outputs[input_step] = self.primitives_outputs[input_step].iloc[indices,:]
            self.primitives_outputs[output_step] = self.primitives_outputs[output_step].iloc[indices,:]
        self.primitives_outputs[dataframe_step] = self.primitives_outputs[dataframe_step].iloc[0,:]
        self.total_cols = len(self.primitives_outputs[input_step].columns) 

    def get_total_cols(self):
        return self.total_cols

    def score_step(self, primitive: PrimitiveBaseMeta, primitive_arguments, metric, posLabel, primitive_desc, hyperparams, step_index, mode):
        """
        Last step of a solution evaluated for score_solution()
        Does hyperparameters tuning
        """
        custom_hyperparams = dict()
        if hyperparams is not None:
            for hyperparam, value in hyperparams.items():
                custom_hyperparams[hyperparam] = value

        if self.cross_outputs is not None:
            return primitive_desc.score_array(self.cross_outputs[0], self.cross_outputs[1], self.cross_outputs[2], self.cross_outputs[3], metric, posLabel, custom_hyperparams, self.taskname, self.id, mode)

        training_arguments_primitive = self._primitive_arguments(primitive, 'set_training_data')
        training_arguments = {}

        for param, value in primitive_arguments.items():
            if param in training_arguments_primitive:
                training_arguments[param] = value
        
        outputs = None
        if 'outputs' in training_arguments:
            outputs = training_arguments['outputs']

        python_path = primitive.metadata.query()['python_path']
        train_data = None
        if 'inputs' in training_arguments:
            train_data = training_arguments['inputs']

        if 'RPI' in python_path:    
            ml_python_path = self.primitives[step_index+2].metadata.query()['python_path'] 
            optimal_params = primitive_desc.optimize_RPI_bins(train_data, outputs, ml_python_path, metric, posLabel)
            score = optimal_params[3]
            return (score, optimal_params)
        else:  
            (score, optimal_params) = primitive_desc.score_primitive(train_data, outputs, metric, posLabel, custom_hyperparams, self.taskname, self.id, mode)
        
        return (score, optimal_params) 

    def describe_solution(self, prim_dict, solution_dict):
        """
        Required for TA2-TA3 API DescribeSolution().
        """
        inputs = []
        for i in range(len(self.inputs)):
            inputs.append(pipeline_pb2.PipelineDescriptionInput(name=self.inputs[i]["name"]))

        outputs=[]
        outputs.append(pipeline_pb2.PipelineDescriptionOutput(name="predictions", data=self.outputs[0][2]))

        steps=[]
        
        for j in range(len(self.primitives_arguments)):
            s = self.primitives[j]
            if s is None: # subpipeline
                s = self.subpipelines[j]
                step_inputs = []
                for argument, data in self.primitives_arguments[j].items():
                    argument_edge = data['data']
                    sa = pipeline_pb2.StepInput(data = argument_edge)
                    step_inputs.append(sa)
                step_outputs = []
                subpipe = solution_dict[s]
                for output in subpipe.outputs:
                    step_outputs.append(pipeline_pb2.StepOutput(id = output[2]))         
                p = pipeline_pb2.SubpipelinePipelineDescriptionStep(pipeline=pipeline_pb2.PipelineDescription(id=s), inputs=step_inputs, outputs=step_outputs)
                steps.append(pipeline_pb2.PipelineDescriptionStep(pipeline=p))
            else: # primitive
                prim = prim_dict[s]
                p = primitive_pb2.Primitive(id=prim.id, version=prim.primitive_class.version, python_path=prim.primitive_class.python_path, name=prim.primitive_class.name, digest=prim.primitive_class.digest)

                arguments={}
                for argument, data in self.primitives_arguments[j].items():
                    argument_edge = data['data']
                    sa = pipeline_pb2.PrimitiveStepArgument(container = pipeline_pb2.ContainerArgument(data=argument_edge))
                    arguments[argument] = sa

                step_outputs = []
                for a in prim.primitive_class.produce_methods:
                    step_outputs.append(pipeline_pb2.StepOutput(id=a))

                hyperparameters = self.get_hyperparams(j,prim_dict)
                step_hyperparameters = {}
                for name, value in hyperparameters.items():
                    if name == "primitive":
                        step_hyperparameters[name] = pipeline_pb2.PrimitiveStepHyperparameter(primitive=pipeline_pb2.PrimitiveArgument(data=value.raw.int64))
                    else:
                        step_hyperparameters[name] = pipeline_pb2.PrimitiveStepHyperparameter(value=pipeline_pb2.ValueArgument(data=value))
                steps.append(pipeline_pb2.PipelineDescriptionStep(primitive=pipeline_pb2.PrimitivePipelineDescriptionStep(primitive=p, arguments=arguments, outputs=step_outputs, hyperparams=step_hyperparameters)))
           
        return pipeline_pb2.PipelineDescription(id=self.id, source=self.source, created=self.created,
         name=self.name, description=self.description, inputs=inputs, outputs=outputs, steps=steps)

    def get_hyperparams(self, step, prim_dict):
        """
        Required for TA2-TA3 API DescribeSolution().
        """
        p = prim_dict[self.primitives[step]]
        custom_hyperparams = self.hyperparams[step]

        send_params={}
       
        if custom_hyperparams is None:
            return send_params
 
        for name, value in custom_hyperparams.items():
            if name == "primitive":
                send_params[name] = value_pb2.Value(raw=value_pb2.ValueRaw(int64=value))
                continue
            else:
                send_params[name] = get_pipeline_values(value)
           
        return send_params



                    
# File: primitivedescription.py 
# Author(s): Saswati Ray
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

import pandas as pd
import numpy as np
import math, sys
from sklearn import metrics, preprocessing
import logging
import d3m.index
from timeit import default_timer as timer
import operator, traceback
import d3m.metadata.base as metadata_base
from d3m import container

from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor, ExtraTreesClassifier, ExtraTreesRegressor, GradientBoostingClassifier, GradientBoostingRegressor
from sklearn.ensemble.bagging import BaggingClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression, Lasso, Ridge
from sklearn.linear_model.stochastic_gradient import SGDClassifier
from sklearn.svm import LinearSVC, LinearSVR, SVC, SVR
from sklearn.model_selection import GridSearchCV

# AutonML imports
import autonml.util as util
from autonml.helper_functions import get_split_indices, get_num_splits

gridsearch_estimators_parameters = {'d3m.primitives.regression.random_forest.SKlearn': [RandomForestRegressor(), 
                                                                                        {'n_estimators': [100],
                                                                                         'max_depth': [8, 10, 15, None],
                                                                                         'min_samples_split': [2, 5, 10]}],
              'd3m.primitives.classification.random_forest.SKlearn': [RandomForestClassifier(),
                                                                      {'n_estimators': [100],
                                                                       'max_depth': [8, 10, 15, None],
                                                                       'min_samples_split': [2, 5, 10],
                                                                       'class_weight': ['balanced', None]}],
              'd3m.primitives.classification.extra_trees.SKlearn': [ExtraTreesClassifier(),
                                                                      {'n_estimators': [100],
                                                                       'max_depth': [8, 10, 15, None],
                                                                       'min_samples_split': [2, 5, 10],
                                                                       'class_weight': ['balanced', None]}],
              'd3m.primitives.regression.extra_trees.SKlearn': [ExtraTreesRegressor(),
                                                                      {'n_estimators': [100],
                                                                       'max_depth': [8, 10, 15, None],
                                                                       'min_samples_split': [2, 5, 10]}],
              'd3m.primitives.classification.gradient_boosting.SKlearn': [GradientBoostingClassifier(),
                                                                          {'n_estimators': [100],
                                                                           'max_depth': [3, 5, 8, 10, 15],
                                                                           'max_features': ['sqrt', None],
                                                                           'min_samples_leaf': [1, 2, 5],
                                                                           'min_samples_split': [2, 5, 10]}],
              'd3m.primitives.regression.gradient_boosting.SKlearn': [GradientBoostingRegressor(),
                                                                      {'n_estimators': [100],
                                                                       'max_depth': [3, 5, 8, 10, 15],
                                                                       'max_features': ['sqrt', None],
                                                                       'min_samples_leaf': [1, 2, 5],
                                                                       'min_samples_split': [2, 5, 10]}],
              'd3m.primitives.classification.linear_svc.SKlearn': [LinearSVC(),
                                                                   {'C': [0.01, 0.1, 1, 10, 100],
                                                                    'class_weight': ['balanced', None]}],
              'd3m.primitives.regression.linear_svr.SKlearn': [LinearSVR(),
                                                                   {'C': [0.01, 0.1, 1, 10, 100]}],
              'd3m.primitives.classification.svc.SKlearn': [SVC(),
                                                            {'C': [0.01, 0.1, 1, 10, 100],
                                                             'class_weight': ['balanced', None]}],
              'd3m.primitives.regression.svr.SKlearn': [SVR(),
                                                        {'C': [0.01, 0.1, 1, 10, 100]}],
              'd3m.primitives.classification.logistic_regression.SKlearn': [LogisticRegression(),
                                                                            {'C': [0.1, 1, 10, 100],
                                                                             'class_weight': ['balanced', None]}],
              'd3m.primitives.regression.ridge.SKlearn': [Ridge(), 
                                                          {'alpha': [0.001, 0.01, 0.1, 1, 5]}],
              'd3m.primitives.regression.lasso.SKlearn': [Lasso(),
                                                          {'alpha': [0.001, 0.01, 0.1, 1, 5]}]
}

def rmse(y_true, y_pred):
    return math.sqrt(metrics.mean_squared_error(y_true, y_pred))

def compute_min_class(y, pos_label):
    frequencies = y.iloc[:,0].value_counts(normalize=True)
    min_freq = frequencies.iat[len(frequencies)-1]
    return min_freq

class PrimitiveDescription(object):
    """
    Class representing single primitive.
    Used for optimizing primitive hyper-parameters, doing cross-validation.
    """
    def __init__(self, primitive, primitive_class):
        self.id = primitive_class.id
        self.primitive = primitive
        self.primitive_class = primitive_class

    def score_ESRNN_primitive(self, prim_instance, X, y):
        """
        ESRNN evaluation! 
        Can only do MAE metric (as yet)!
        Retrieve internal CV score
        """
        prim_instance.set_training_data(inputs=X, outputs=y)
        prim_instance.fit()
        try:
            metric = min(prim_instance._esrnn.losses)
        except:
            metric = 0

        return metric

    def create_optimal_parameters(self, X):
        optimal_params = dict()
        python_path = self.primitive.metadata.query()['python_path']

        # Put constraints on primitive hyperparameters to avoid excessive time-complexity!
        if 'SKlearn' in python_path:
            hyperparam_spec = self.primitive.metadata.query()['primitive_code']['hyperparams']
            if len(X) >= 50000 and ('random_forest' in python_path or 'gradient_boosting' in python_path or 'bagging' in python_path): # 'n_estimators' in hyperparam_spec:
                optimal_params['n_estimators'] = 50
            if len(X) >= 100000 and 'n_estimators' in hyperparam_spec:
                optimal_params['n_estimators'] = 10
            if (len(X) >= 100000 or len(X.columns) >= 5000) and ('linear_svc' in python_path or 'linear_svr' in python_path):
                optimal_params['max_iter'] = 100
            if len(X.columns) > 500 and ('bagging' in python_path):
                optimal_params['max_features'] = 15
            if len(X.columns) > 50 and 'gradient_boosting' in python_path:
                optimal_params['max_features'] = 7
            if len(X.columns) >= 5000 and ('random_forest' in python_path or 'extra_trees' in python_path):
                optimal_params['max_features'] = 30
        return optimal_params 
 
    def score_primitive(self, X, y, metric_type, posLabel, custom_hyperparams, taskname, pipeline_id=None, mode=None):
        """
        Learns optimal hyperparameters for the primitive
        Returns metric score and optimal parameters.
        X: Training data inputs
        y: Training data outputs
        metric_type: Scoring metric
        posLabel: Positive label for F1 metric
        """
        python_path = self.primitive.metadata.query()['python_path']

        optimal_params = self.create_optimal_parameters(X)

        if custom_hyperparams is not None:
            for name, value in custom_hyperparams.items():
                optimal_params[name] = value

        # No CV for these primitives!
        if X is None or y is None or 'graph' in python_path or 'Vertex' in python_path or \
            'DistilLink' in python_path or \
            'community' in python_path or \
            'JHU' in python_path or \
            'yolo' in python_path or \
            'FCN' in python_path or \
            'retina' in python_path:
            if util.invert_metric(metric_type) is True:
                return (0.0, optimal_params)
            else:
                if 'JHU' in python_path:
                    return (0.99, optimal_params)
                if 'retina' in python_path:
                    return (0.99, optimal_params)
                return (1.0, optimal_params)

        # Lasso CV can become very expensive for large number of columns!!!
        # Use lasso's CV score
        if 'lasso_cv' in python_path and len(X.columns) > 500:
            return (1.0, optimal_params)

        primitive_hyperparams = self.primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        prim_instance = self.primitive(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **optimal_params))
        score = 0.0

        # ESRNN evaluation! 
        # Can only do MAE metric (as yet)!
        # Retrieve internal CV score
        if 'esrnn' in python_path:
            start = timer()
            score = self.score_ESRNN_primitive(prim_instance, X, y)
            end = timer()
            if mode is not None:
                try:
                    predictions = prim_instance.produce(X)
                    outputFilePath = mode + "/" + str(pipeline_id) + "_train_predictions.csv"
                    with open(outputFilePath, 'w') as outputFile:
                        predictions.to_csv(outputFile, header=True)
                except:
                    predictions = None
                    
            logging.critical("Time taken for esrnn = %s secs", end-start) 
            logging.critical("MAE Score for %s = %s", python_path, score)
            return (score, optimal_params)

        splits = get_num_splits(len(X), len(X.columns))
        
        # Hard coding MAE for forecasting problems since thats what esrnn supports!
        if taskname == 'FORE_REGRESSION' or 'FORECASTING' in taskname:
            metric_type = "MEAN_ABSOLUTE_ERROR"

        logging.critical("%s - %s - %s", python_path, X.shape, optimal_params)
        # Run k-fold CV and compute mean metric score 
        (score, metric_scores) = self.k_fold_CV(prim_instance, X, y, metric_type, posLabel, splits, pipeline_id, mode)
        return (score, optimal_params)

    def score_array(self, X, X2, y, y2, metric_type, posLabel, custom_hyperparams, taskname, pipeline_id=None, mode=None):
        """
        Score a primitive using given train-test splits
        X: Train splits
        X2: Test splits
        y: Train output splits
        y2: Test output splits
        """
        python_path = self.primitive.metadata.query()['python_path']
        optimal_params = self.create_optimal_parameters(X[0])

        if custom_hyperparams is not None:
            for name, value in custom_hyperparams.items():
                optimal_params[name] = value

        if 'lasso_cv' in python_path and len(X[0].columns) > 500:
            return (1.0, optimal_params)

        primitive_hyperparams = self.primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        prim_instance = self.primitive(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **optimal_params))
        score = 0.0

        for X_train, y_train, X_test, y_test in zip(X, y, X2, y2):
            prim_instance.set_training_data(inputs=X_train, outputs=y_train)
            prim_instance.fit()
            predictions = prim_instance.produce(inputs=X_test).value
            if 'xgboost' in python_path and len(predictions.columns) > 1:
                predictions = predictions.iloc[:,len(predictions.columns)-1]
            metric = self.evaluate_metric(predictions, y_test, metric_type, posLabel)
            score += metric
        logging.critical("%s - %s - %s", python_path, X[0].shape, optimal_params)
        score /= len(X)
        return (score, optimal_params)

    def evaluate_metric(self, predictions, Ytest, metric, posLabel):
        """
        Function to compute metric score for predicted-vs-true output.
        """
        count = len(Ytest)

        if metric == "ACCURACY":
            return metrics.accuracy_score(Ytest, predictions)
        elif metric == "PRECISION":
            return metrics.precision_score(Ytest, predictions)
        elif metric == "RECALL":
            return metrics.recall_score(Ytest, predictions)
        elif metric == "F1":
            return metrics.f1_score(Ytest, predictions, pos_label=posLabel)
        elif metric == "F1_MICRO":
            return metrics.f1_score(Ytest, predictions, average='micro')
        elif metric == "F1_MACRO":
            return metrics.f1_score(Ytest, predictions, average='macro')
        elif metric == "ROC_AUC":
            return metrics.roc_auc_score(Ytest, predictions, multi_class='ovr')
        elif metric == "ROC_AUC_MICRO":
            return metrics.roc_auc_score(Ytest, predictions, multi_class='ovr', average='micro')
        elif metric == "ROC_AUC_MACRO":
            return metrics.roc_auc_score(Ytest, predictions, multi_class='ovr', average='macro')
        elif metric == "MEAN_SQUARED_ERROR":
            return metrics.mean_squared_error(Ytest, predictions)
        elif metric == "ROOT_MEAN_SQUARED_ERROR":
            return math.sqrt(metrics.mean_squared_error(Ytest, predictions))
        elif metric == "MEAN_ABSOLUTE_ERROR":
            return metrics.mean_absolute_error(Ytest, predictions)
        elif metric == "R_SQUARED":
            return metrics.r2_score(Ytest, predictions)
        elif metric == "NORMALIZED_MUTUAL_INFORMATION":
            return metrics.normalized_mutual_info_score(Ytest, predictions)
        elif metric == "JACCARD_SIMILARITY_SCORE":
            return metrics.jaccard_similarity_score(Ytest, predictions)
        elif metric == "PRECISION_AT_TOP_K":
            return 0.0
        elif metric == "OBJECT_DETECTION_AVERAGE_PRECISION":
            return 0.0
        elif metric == "MEAN_RECIPROCAL_RANK":
            return 0.0
        elif metric == "HITS_AT_K":
            return 0.0
        else:
            return metrics.accuracy_score(Ytest, predictions)

    def optimize_primitive_gridsearch(self, train, output, python_path, metric, posLabel):
        # Do grid-search to learn optimal parameters for the model
        if python_path in gridsearch_estimators_parameters:
            start = timer()
            (model, search_grid) = gridsearch_estimators_parameters[python_path]
            splits = get_num_splits(len(train), len(train.columns))

            from sklearn.metrics import make_scorer
            if metric == "ACCURACY":
                scorer = make_scorer(metrics.accuracy_score)
            elif metric == "F1":
                scorer = make_scorer(metrics.f1_score, pos_label=posLabel)
            elif metric == "F1_MACRO":
                scorer = make_scorer(metrics.f1_score, average='macro')
            elif metric == "MEAN_SQUARED_ERROR":
                scorer = make_scorer(metrics.mean_squared_error, greater_is_better=False)
            elif metric == "ROOT_MEAN_SQUARED_ERROR": 
                scorer = make_scorer(rmse, greater_is_better=False)
            else:
                scorer = None

            rf_random = GridSearchCV(estimator = model, param_grid = search_grid, scoring = scorer, cv = splits, verbose=0, n_jobs = -1)

            # Fit the random search model
            rf_random.fit(train, output)
            print(rf_random.best_params_)
            end = timer() 
            print("Time taken for ", python_path, " = ", end-start, " secs")
            return rf_random.best_params_
        else:
            print("No grid search done for ", python_path)
            return None

    def optimize_F1_metric(self, train, y, optimal_params, metric_type, posLabel):
        corex_hp = {'class_weight': ['balanced', None]}

        RPI_path = self.primitive.metadata.query()['python_path']
        primitive_hyperparams = self.primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']

        splits = get_num_splits(len(train), len(train.columns))
        scores = {}

        start = timer()
        for i in corex_hp['class_weight']:
            optimal_params['class_weight'] = i
            model = self.primitive(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **optimal_params))
            (score, metric_scores) = self.k_fold_CV(model, train, y, metric_type, posLabel, splits)
            mean = np.mean(metric_scores)
            scores[(i,mean)] = mean
        print(RPI_path)
        print(scores)
        sorted_x = sorted(scores.items(), key=operator.itemgetter(1))
        sorted_x.reverse()
        (key, value) = sorted_x[0]
        end = timer()
        logging.critical("%s time taken = %s secs", RPI_path, end-start)
        return key

    def optimize_RPI_bins(self, train, y, python_path, metric_type, posLabel):
        corex_hp = {'nbins': [2, 3, 4, 10],# 12, 15, 20],
                    'method': ['counting', 'fullBayesian'], #'pseudoBayesian'],
                    'n_estimators': [20, 25, 30, 32]}
   
        RPI_path = self.primitive.metadata.query()['python_path'] 
        # For simultaneous_markov_blanket.AutoRPI
        if 'markov' in RPI_path:
            corex_hp['nbins'] = [10]
            corex_hp['method'] = ['counting']#, 'pseudoBayesian', 'fullBayesian', 'BayesFactor']

        if(len(train) > 25000): # Use defaults. No tuning
            corex_hp = {'nbins': [10],
                        'method': ['counting'],
                        'n_estimators': [10]}

        if 'linear_discriminant_analysis' in python_path:
            corex_hp['n_estimators'] = [0]

        prim = d3m.index.get_primitive(python_path)
        model_hyperparams = prim.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']

        primitive_hyperparams = self.primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']

        sklearn_prim = d3m.index.get_primitive('d3m.primitives.data_cleaning.imputer.SKlearn')
        sklearn_hyperparams = sklearn_prim.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        custom_hyperparams = dict()
        custom_hyperparams['strategy'] = 'most_frequent'
        sklearn_primitive = sklearn_prim(hyperparams=sklearn_hyperparams(sklearn_hyperparams.defaults(), **custom_hyperparams))

        splits = get_num_splits(len(train), len(train.columns))
        scores = {}

        start = timer()

        for i in corex_hp['nbins']:
            custom_hyperparams = dict()
            custom_hyperparams['nbins'] = i

            for j in corex_hp['method']:
                custom_hyperparams['method'] = j
                model = self.primitive(hyperparams=primitive_hyperparams(primitive_hyperparams.defaults(), **custom_hyperparams))

                model.set_training_data(inputs=train, outputs=y)
                model.fit()
                output = model.produce(inputs=train).value

                sklearn_primitive.set_training_data(inputs=output)
                sklearn_primitive.fit()
                output = sklearn_primitive.produce(inputs=output).value

                for k in corex_hp['n_estimators']:
                    model_hp = dict()
                    model_hp['n_estimators'] = k
                    if 'gradient_boosting' in python_path:
                        model_hp['learning_rate'] = 10/k
                    if 'linear_discriminant_analysis' in python_path:
                        rf_model = prim(hyperparams=model_hyperparams(model_hyperparams.defaults()))
                    else:
                        rf_model = prim(hyperparams=model_hyperparams(model_hyperparams.defaults(), **model_hp))
                    (score, metric_scores) = self.k_fold_CV(rf_model, output, y, metric_type, posLabel, splits)
                    mean = np.mean(metric_scores)
                    median = np.median(metric_scores)
                    stderror = np.std(metric_scores)/math.sqrt(len(metric_scores))
                    z = 1.96*stderror
                    scores[(i,j,k,mean)] = mean#-z

        sorted_x = sorted(scores.items(), key=operator.itemgetter(1))
        if util.invert_metric(metric_type) is False:
            sorted_x.reverse()
        (key, value) = sorted_x[0]
        end = timer()
        logging.critical("%s time taken for %s = %s secs", RPI_path, python_path, end-start)
        return key

    def k_fold_CV(self, prim_instance, X, y, metric_type, posLabel, splits, pipeline_id=None, mode=None):
        """
        Run k-fold CV.
        k = splits
        prim_instance has already been initialized with hyperparameters.
        """
        python_path = self.primitive.metadata.query()['python_path']
        metric_sum = 0
        score = 0.0

        if 'NBEATS' in python_path:
            metric = 5000
            return (metric, [metric, metric])
        if 'VAR' in python_path or 'arima' in python_path or 'nbeats' in python_path:
            start = timer()
            prim_instance.set_training_data(inputs=X, outputs=y)
            prim_instance.fit()
            predictions = prim_instance.produce(inputs=X).value
            predictions = predictions.iloc[:,len(predictions.columns)-1]
            if mode is not None:
                outputFilePath = mode + "/" + str(pipeline_id) + "_train_predictions.csv"
                with open(outputFilePath, 'w') as outputFile:
                    predictions.to_csv(outputFile, header=True)
            metric = self.evaluate_metric(predictions, y, metric_type, posLabel)
            end = timer()
            logging.critical("Time taken for %s = %s secs (%f)", python_path, end-start, metric)
            return (metric, [metric, metric])

        # Run k-fold CV and compute mean metric score
        metric_scores = []
        split_indices = get_split_indices(X, y, splits, python_path)

        start = timer()
        train_predictions = None

        if mode is not None:
            # Create training predictions
            train_predictions = pd.DataFrame(columns=["Prediction"], index=np.arange(len(X)))
            if "ROC_AUC" in metric_type:
                labels = sorted(y.iloc[:,0].unique())
                if labels[0] == '':
                    labels = labels[1:]
                num_classes = len(labels) 
                train_predictions = pd.DataFrame(columns=["Prediction"], index=np.repeat(np.arange(len(X)),num_classes))

        # Do the actual k-fold CV here
        for train_index, test_index in split_indices:
            X_train, X_test = X.iloc[train_index,:], X.iloc[test_index,:]
            y_train, y_test = y.iloc[train_index,:], y.iloc[test_index,:]

            X_test.reset_index(drop=True,inplace=True)
            prim_instance.set_training_data(inputs=X_train, outputs=y_train)
            prim_instance.fit()

            # Compute predictions on test data
            if "ROC_AUC" in metric_type:
                input_labels = container.DataFrame(labels)
                input_labels = input_labels.append([input_labels] * (len(y_test) - 1)).reset_index(drop=True)
                input_labels.metadata = input_labels.metadata.generate(input_labels)
                input_labels.metadata = input_labels.metadata.add_semantic_type((metadata_base.ALL_ELEMENTS, 0), 'https://metadata.datadrivendiscovery.org/types/TrueTarget')
                primitive_inputs = X_test.loc[X_test.index.repeat(len(labels))].reset_index(drop=True)
                primitive_inputs.metadata = X_test.metadata.update((), {'dimension': {'length': len(labels)}})
                predictions = prim_instance.log_likelihoods(inputs=primitive_inputs,outputs=input_labels).value
                predictions = predictions.apply(np.vectorize(np.exp))
            else:
                predictions = prim_instance.produce(inputs=X_test).value

            if ('xgboost' in python_path or 'DeepAR' in python_path) and len(predictions.columns) > 1:
                predictions = predictions.iloc[:,len(predictions.columns)-1]

            # Compute evaluation metric on test set 
            try:
                if "ROC_AUC" in metric_type:
                    indices = [np.arange(i*num_classes, i*num_classes+num_classes) for i in test_index]
                    indices = np.ravel(indices)
                    train_predictions.iloc[indices] = predictions.values
                    output = predictions.groupby(np.arange(len(predictions)) // num_classes, as_index=False).aggregate(lambda x: tuple(v for v in x))
                    output.reset_index(inplace=True)
                    output.drop(columns='index',inplace=True)
                    encoded = output.iloc[:,0].apply(pd.Series)
                    encoded.columns = labels
                    if len(labels) == 2:
                        encoded = encoded.iloc[:,1]
                    predictions = encoded.to_numpy()
                    y_test = np.squeeze(y_test.values)
                    if 'iterative_labeling' in python_path: # Semi-supervised ?
                        labeledIx = np.where(y_test[:] != '')[0]
                        predictions = predictions[labeledIx]
                        y_test = y_test[labeledIx]
                elif 'iterative_labeling' in python_path: # Semi-supervised ?
                    labeledIx = np.where(y_test.iloc[:, 0].values != '')[0]
                    predictions = predictions.iloc[labeledIx]
                    y_test = y_test.iloc[labeledIx]  
                metric = self.evaluate_metric(predictions, y_test, metric_type, posLabel)
            except:
                metric = 0

            if train_predictions is not None and "ROC_AUC" not in metric_type:
                if 'xgboost' in python_path:
                    train_predictions.loc[test_index, :] = predictions.values.reshape(-1, 1)
                else:
                    train_predictions.loc[test_index, :] = predictions.values

            metric_scores.append(metric)
            metric_sum += metric

        if train_predictions is not None:
            outputFilePath = mode + "/" + str(pipeline_id) + "_train_predictions.csv"
            with open(outputFilePath, 'w') as outputFile:
                train_predictions.to_csv(outputFile, header=True)

        score = metric_sum/splits
        end = timer()
        if 'RPI' not in python_path:
            logging.critical("Time taken for %s = %s secs (%f)", python_path, end-start, score)
        return (score, metric_scores)




import os, time, random, copy, logging, typing, math, subprocess
from enum import Enum
from google.protobuf.timestamp_pb2 import Timestamp
import numpy as np
import pandas as pd
import d3m.index
from timeit import default_timer as timer
from d3m.metadata import base as metadata_base

# AutonML imports
import autonml.value_pb2 as value_pb2

def get_argument(data):
    """
    Split string into origin and source for pipeline construction
    """
    origin = data.split('.')[0]
    source = data.split('.')[1]
    return {'origin': origin, 'source': int(source), 'data': data}

def isFloat(value):
    """
    Is a float?
    """
    try:
        a  = float(value)
        return True
    except:
        return False

def isOrdinal(df, t):
    """
    Whether a categorical attribute needs to be treated as ordinal?
    """
    missing = 0
    if df.dtypes[t] == 'object':
        missing = len(np.where(df.iloc[:,t] == '')[0])
        if missing == 0:
            try:
                pd.to_numeric(df.iloc[:,t])
                return True
            except:
                print("Att ", df.columns[t], " non-numeric")
    return False

def isNumeric(df, t):
    """
    Is a numeric attribute?
    """
    anynumber = False
    try:
        for i in range(5):
            if isFloat(df.iloc[i,t]) is True:
                anynumber = True
                break
    except:
        print("Att ", df.columns[t], " non-numeric")
    return anynumber

def compute_timestamp():
    now = time.time()
    seconds = int(now)
    return Timestamp(seconds=seconds)

class StepType(Enum):
    PRIMITIVE = 1
    SUBPIPELINE = 2
    PLACEHOLDER = 3

class ActionType(Enum):
    FIT = 1
    SCORE = 2
    VALIDATE = 3

def get_values(arg):
    value = None
    if arg.HasField("double") == True:
        value = arg.double
    elif arg.HasField("int64") == True:
        value = arg.int64
    elif arg.HasField("bool") == True:
        value = arg.bool
    elif arg.HasField("string") == True:
        value = arg.string
    elif arg.HasField("bytes") == True:
        value = arg.bytes
    elif arg.HasField("list") == True:
        value = get_list_items(arg.list.items)
    elif arg.HasField("dict") == True:
        value = get_dict_items(arg.dict.items)

    return value

def get_list_items(values : value_pb2.ValueList):
    items = []

    for i in values:
        items.append(get_values(i))

    return items

def get_dict_items(values : value_pb2.ValueDict):
    items = {}

    for name, arg in values.items():
        items[name] = get_values(arg)
    return items

def get_pipeline_values(value):
    if value is None:
        return value_pb2.Value(raw=value_pb2.ValueRaw(null=value_pb2.NULL_VALUE))
    if isinstance(value, list):
        vitems = []
        for v in value:
            vitems.append(get_pipeline_values(v).raw)
        vlist = value_pb2.ValueList(items=vitems)
        return value_pb2.Value(raw=value_pb2.ValueRaw(list=vlist))
    elif isinstance(value, dict):
        vitems = {}
        for name,v in value.items():
            vitems[name] = get_pipeline_values(v).raw
        vlist = value_pb2.ValueDict(items=vitems)
        return value_pb2.Value(raw=value_pb2.ValueRaw(dict=vlist))
    elif isinstance(value, bool):
        return value_pb2.Value(raw=value_pb2.ValueRaw(bool=value))
    elif isinstance(value, int):
        return value_pb2.Value(raw=value_pb2.ValueRaw(int64=value))
    elif isinstance(value, float):
        return value_pb2.Value(raw=value_pb2.ValueRaw(double=value))
    elif isinstance(value, str):
        return value_pb2.Value(raw=value_pb2.ValueRaw(string=value))
    return None

def get_cols_to_encode(df):
    """
    Find categorical attributes which can be one-hot-encoded.
    Find categorical attributes which can be treated as ordinals.
    Find attributes of UnknownType which can be treated as floats.
    """
    cols = df.metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/CategoricalData")
    targets = df.metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/TrueTarget")
    attributes = df.metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/Attribute")

    for t in targets:
        if t in cols:
            cols.remove(t)

    rows = len(df)
    # use rule of thumb to exclude categorical atts with high cardinality for one-hot-encoding
    max_num_cols = math.log(rows, 2)
    if len(attributes) >= 100 and max_num_cols > 0.05*len(attributes):
        max_num_cols = 0.05*len(attributes)

    if rows > 100000:
        max_num_cols = max_num_cols/4

    tmp_cols = copy.deepcopy(cols)
    ordinals = []

    # Iterate over all categorical attributes
    for t in tmp_cols:
        if isFloat(t) == False:
            cols.remove(t)
            continue

        arity = len(df.iloc[:,t].unique())
        if arity == 1:
            cols.remove(t)
            continue

        if arity > max_num_cols:
            cols.remove(t)
            if isOrdinal(df, t) is True:
                ordinals.append(t)
        elif rows > 1000000 and isOrdinal(df, t) is True:
            cols.remove(t)
            ordinals.append(t)

    if rows > 100000 and len(cols) > 5:
        cols = random.sample(cols, 5)

    add_floats = []
    for att in attributes:
        md = df.metadata.query((metadata_base.ALL_ELEMENTS, att))
        if md['structural_type'] != str:
            continue

        if 'semantic_types' not in md:
            continue

        attmeta = md['semantic_types']
        length = len(attmeta)-1
        if 'https://metadata.datadrivendiscovery.org/types/UniqueKey' in attmeta:
            continue
        if 'https://metadata.datadrivendiscovery.org/types/UnknownType' in attmeta:
            length = length-1
        if length == 0:
            anynumber = isNumeric(df, att)
            if anynumber is True:
                logging.critical("Float for %s", att)
                add_floats.append(int(att))
    logging.info("No. of cats = %s", len(cols))
    logging.info("Floats = %s", add_floats)
    return (list(cols), ordinals, add_floats)

def get_primitive_volumes(volumes_dir, primitive_class, primitive_name="") -> typing.Dict:
        volumes = {}
        for entry in primitive_class.metadata.get_volumes():
            volume_path = os.path.join(volumes_dir, entry['file_digest'])
            if not os.path.exists(volume_path):
                subprocess.run(["python", "-m", "d3m", "primitive", "download", "-p", primitive_name, "-o", os.environ['D3MSTATICDIR']])
            volumes[entry['key']] = volume_path
        return volumes

def get_num_splits(length, cols):
    splits = 2
    if length < 500:
        splits = 50
        if length < splits:
            splits = length
    elif length < 1000:
        splits = 25
    elif length < 2500:
        splits = 20
    elif length < 5000:
        splits = 10
    elif length < 10000:
        splits = 5
    elif length < 20000:
        splits = 3
    else:
        splits = 2

    #if cols > 500:
    splits = min(5, splits)
    return splits

def get_split_indices(X, y, splits, python_path='classification'):
    if 'classification' in python_path: # Classification
        frequencies = y.iloc[:,0].value_counts()
        min_freq = frequencies.iat[len(frequencies)-1]
        if min_freq < splits:
            from sklearn.model_selection import KFold as KFold
            kf = KFold(n_splits=splits, shuffle=True, random_state=9001)
            split_indices = kf.split(X)
        else:
            from sklearn.model_selection import StratifiedKFold as KFold
            kf = KFold(n_splits=splits, shuffle=True, random_state=9001)
            split_indices = kf.split(X, y)
    else: # Regression
        from sklearn.model_selection import KFold as KFold
        kf = KFold(n_splits=splits, shuffle=True, random_state=9001)
        split_indices = kf.split(X)
    return split_indices

def column_types_present(dataset):
    """
    Retrieve special data types present: Text, Image, Timeseries, Audio, Categorical
    Returns ([data types], total columns, total rows, [categorical att indices], ok_to_denormalize, privileged, add_floats)
    """
    ok_to_denormalize = True
    start = timer()

    try:
        primitive = d3m.index.get_primitive('d3m.primitives.data_transformation.denormalize.Common')
        primitive_hyperparams = primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
        model = primitive(hyperparams=primitive_hyperparams.defaults())
        ds = model.produce(inputs=dataset).value
        dataset = ds
    except:
        print("Exception with denormalize!")
        ok_to_denormalize = False

    primitive = d3m.index.get_primitive('d3m.primitives.data_transformation.dataset_to_dataframe.Common')
    primitive_hyperparams = primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
    model = primitive(hyperparams=primitive_hyperparams.defaults())
    df = model.produce(inputs=dataset).value
    atts = df.metadata.get_columns_with_semantic_type('https://metadata.datadrivendiscovery.org/types/UnknownType')
    print("Dataset shape = ", len(df), ", ", len(df.columns))

    if len(df.columns) > 5000:
        return ([], len(df.columns), len(df), [], [], True, [], [])

    profiler_needed = False
    primitive = d3m.index.get_primitive('d3m.primitives.schema_discovery.profiler.Common')
    primitive_hyperparams = primitive.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
    model = primitive(hyperparams=primitive_hyperparams.defaults())
    model.set_training_data(inputs=df)
    try:
        model.fit()
        df = model.produce(inputs=df).value
    except:
        profiler_needed = False

    metadata = df.metadata

    types = []
    (categoricals, ordinals, add_floats) = get_cols_to_encode(df)
    if len(categoricals) > 0:
        types.append('Categorical')
        logging.critical("Cats = %s", categoricals)
    if len(ordinals) > 0:
        types.append('Ordinals')
        print("Ordinals = ", ordinals)

    textcols = metadata.get_columns_with_semantic_type("http://schema.org/Text")
    if len(textcols) > 0:
        logging.critical("Text = %s", textcols)
        types.append('COREXTEXT')
    cols = len(metadata.get_columns_with_semantic_type("http://schema.org/ImageObject"))
    if cols > 0:
        types.append('IMAGE')
    cols = len(metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/Timeseries"))
    if cols > 0:
        types.append('TIMESERIES')
    cols = len(metadata.get_columns_with_semantic_type("http://schema.org/AudioObject"))
    if cols > 0:
        types.append('AUDIO')
    cols = len(metadata.get_columns_with_semantic_type("http://schema.org/VideoObject"))
    if cols > 0:
        types.append('VIDEO')
    cols = len(metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/FileName"))
    if cols > 0:
        types.append('FILES')

    privileged = metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/PrivilegedData")
    attcols = metadata.get_columns_with_semantic_type("https://metadata.datadrivendiscovery.org/types/Attribute")

    print("Data types present: ", types)
    end = timer()
    logging.critical("Time taken in determining data types: %f secs", (end-start))
    return (types, len(attcols), len(df), categoricals, ordinals, ok_to_denormalize, privileged, add_floats)




# File: __init__.py 
# Author(s): Vedant Sanil
# Created: Wed Feb 17 11:49:20 EST 2022 
# Description:
# Acknowledgements:
# Copyright (c) 2022 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

"""AutonML package"""

__VERSION_MAJOR__  = '0'
__VERSION_MINOR__ = '1'
__VERSION_BUILD__ = '8'

__VERSION__ = f'{__VERSION_MAJOR__}.{__VERSION_MINOR__}.{__VERSION_BUILD__}'

from autonml.autonml_api import AutonML



# File: main.py 
# Author(s): Saswati Ray, Vedant Sanil
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

from ctypes import ArgumentError
import logging
from multiprocessing import set_start_method
set_start_method("spawn", force=True)
from concurrent import futures
import time
import sys
import os
import json
from pathlib import Path
import shutil
import warnings
import subprocess

warnings.filterwarnings("ignore")
sys.path.append(f'{os.getcwd()}/autonml/')

import grpc
from multiprocessing import cpu_count

# AutonML imports
import autonml.search as search
from autonml.api_v3 import core

TA2_API_HOST = '[::]'
TA2_API_PORT = 45042

def garbage_collect():

    # TODO (vedant) : garbage collection for NBeats is being done
    # for kungfuai primitives. This needs to be resolved within
    # kungfuai NBeats
    if os.path.exists(os.path.join(os.getcwd(), "nbeats_weights")):
        shutil.rmtree(os.path.join(os.getcwd(), "nbeats_weights"))

def prerun_error_check():
    if len(sys.argv) < 6 or len(sys.argv) > 7:
        arg_err_str = "Incorrect number of arguments provided. Required arguments: \n\t-Input Directory \n\t-Output Directory \n\t-Timeout(in minutes) \n\t-Number of CPU cores \n\t-D3M Problem Path (please refer to AutonML's documentation for more details \n\n Optional arguments \n\t-D3M Static directory: stores weight files\n"
        raise ArgumentError(arg_err_str)

    if not os.path.exists(sys.argv[1]):
        raise FileNotFoundError("Input directory does not exist")

    if not os.path.exists(sys.argv[5]):
        raise FileNotFoundError("Problem Path directory does not exist")

    if len(sys.argv) == 7:
        if not os.path.exists(sys.argv[6]):
            raise FileNotFoundError("Static directory does not exist")

def main_run():
    '''Main API to invoke complete AutoML pipeline'''

    if not os.path.exists(sys.argv[2]):
        logging.warning(f"Output directory does not exist, creating Output directory at path: {output_path}")
    
    os.makedirs(sys.argv[2], exist_ok=True)

    # Perform a preliminary error check to ensure number of arguments and directories supplied are correct
    prerun_error_check()

    # Remove all files in output directory
    output_dir = sys.argv[2]
    for search_dir in os.listdir(output_dir):
        search_dir_path = os.path.join(output_dir, search_dir)
        shutil.rmtree(search_dir_path)

    # Search is invoked first to create optimal pipelines
    main_search()

    # Fit and evaluate across searched pipelines
    # TODO (vedant) : error checks here
    fit_and_evaluate(output_dir)

    # Save environment variables in a file
    for search_dir in os.listdir(sys.argv[2]):
        search_dir_path = os.path.join(sys.argv[2], search_dir)
        filepath = os.path.join(search_dir_path, 'environment_variables.json')
        
        env_dict = {}
        env_dict['D3MINPUTDIR'] = os.environ['D3MINPUTDIR']
        env_dict['D3MSTATICDIR'] = os.environ['D3MSTATICDIR']
        env_dict['D3MOUTPUTDIR'] = os.environ['D3MOUTPUTDIR']
        env_dict['D3MTIMEOUT'] = os.environ['D3MTIMEOUT']
        env_dict['D3MCPU'] = (int)(os.environ['D3MCPU'])

        with open(filepath, "w") as fp:
            json.dump(env_dict, fp)

    # Garbage collect any leftover intermediate variables
    garbage_collect()


def fit_and_evaluate(output_dir):

    # TODO (vedant) : Error checking on valid directory paths?
    # Code to score the output, and write to the $D3MOUTPUTDIR directory
    for search_dir in os.listdir(output_dir):
        search_dir_path = os.path.join(output_dir, search_dir)
        pipeline_dir_path = os.path.join(search_dir_path, 'pipelines_ranked')

        for idx, jsonfile in enumerate(os.listdir(pipeline_dir_path)):
            jsonfile_path = os.path.join(pipeline_dir_path, jsonfile)
            prediction_name = jsonfile.replace('.json', '')
            logging.critical(f'Evaluating over JSON File {idx+1}/{len(os.listdir(pipeline_dir_path))}: {jsonfile_path}\n')

            p = subprocess.Popen([sys.executable, '-m', 'd3m', 'runtime', '-v',
                                  os.environ['D3MSTATICDIR'], 'fit-produce', '-p',
                                  jsonfile_path, '-r', os.path.join(os.environ['D3MINPUTDIR'], 'TRAIN/problem_TRAIN/problemDoc.json'),
                                  '-i', os.path.join(os.environ['D3MINPUTDIR'], 'TRAIN/dataset_TRAIN/datasetDoc.json'), '-t',
                                  os.path.join(os.environ['D3MINPUTDIR'], 'TEST/dataset_TEST/datasetDoc.json'), '-o',
                                  os.path.join(search_dir_path, 'predictions', f'{prediction_name}.predictions.csv')], 
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            output, error = p.communicate()
            if p.returncode != 0:
                print(output.decode())
                raise RuntimeError(error.decode())

            # Generate code that converts JSON to python code
            src_dir = os.path.dirname(os.path.abspath(__file__))

            subprocess.run([sys.executable, os.path.join(src_dir, 'mkpline.py'), 
                                jsonfile_path, os.path.join(search_dir_path, 'executables', f'{prediction_name}.code.py')],
                                stdout=sys.stdout, stderr=sys.stdout, check=True)


def main_search():
    '''API called only during search'''
    # TODO (vedant) : error handling if number of args provided is less than zero? 
    # is this very necessary ?

    # TODO (vedant) : pass D3m specific variables directly instead of environment 
    # variables
    logging.info('RUNNING SEARCH')

    os.environ['D3MINPUTDIR'] = sys.argv[1]
    os.environ['D3MOUTPUTDIR'] = sys.argv[2]
    os.environ['D3MTIMEOUT'] = sys.argv[3]
    os.environ['D3MCPU'] = sys.argv[4]
    os.environ['D3MPROBLEMPATH'] = sys.argv[5]

    if len(sys.argv) == 7:
        os.environ['D3MSTATICDIR'] = sys.argv[6]
    else:
        os.environ['D3MSTATICDIR'] = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'static')

    search.search_phase()

def main(argv):
    mode = argv[0]
    logging.info("Running in mode %s", mode)

    if mode == "search":
        search.search_phase()
    else:
        threadpool = futures.ThreadPoolExecutor(max_workers=cpu_count())
        server = grpc.server(threadpool)
        core.add_to_server(server)
        server_string = '{}:{}'.format(TA2_API_HOST, TA2_API_PORT)
        server.add_insecure_port(server_string)
        logging.critical("Starting server on %s", server_string)
        server.start()
        logging.critical("Server started, waiting.")
        try:
            while True:
                time.sleep(3600)
        except KeyboardInterrupt:
            server.stop(0)

if __name__ == '__main__':
    #main(sys.argv[1:])
    main_run()




# File: util.py 
# Author(s): Saswati Ray
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

import logging, uuid

__version__ = "0.1.0"

from d3m.container.dataset import D3MDatasetLoader, Dataset
from d3m.metadata import base as metadata_base, problem
from d3m.metadata.base import Metadata
from d3m.metadata.pipeline import Pipeline, PrimitiveStep
import os, json
import pickle
import pandas as pd
import numpy as np

# AutonML imports
import autonml.solutiondescription as solutiondescription
import autonml.problem_pb2 as problem_pb2

def load_problem_doc(problem_doc_uri: str):
    """     
    Load problem_doc from problem_doc_uri     
    Parameters     ---------     
    problem_doc_uri     Uri where the problemDoc.json is located
    """
    with open(problem_doc_uri) as file:
        problem_doc = json.load(file)
    problem_doc_metadata = Metadata(problem_doc)
    return problem_doc_metadata

def add_target_columns_metadata(dataset: 'Dataset', problem_doc: 'Metadata'):

    for data in problem_doc['inputs']:
        targets = data['targets']
        for target in targets:
            semantic_types = list(dataset.metadata.query((target['resource_id'], metadata_base.ALL_ELEMENTS, target['column_index'])).get('semantic_types', []))
            if 'https://metadata.datadrivendiscovery.org/types/Target' not in semantic_types:
                semantic_types.append('https://metadata.datadrivendiscovery.org/types/Target')
                dataset.metadata = dataset.metadata.update((target['resource_id'], metadata_base.ALL_ELEMENTS, target['column_index']), {'semantic_types': semantic_types})
            if 'https://metadata.datadrivendiscovery.org/types/TrueTarget' not in semantic_types:
                semantic_types.append('https://metadata.datadrivendiscovery.org/types/TrueTarget')
                dataset.metadata = dataset.metadata.update((target['resource_id'], metadata_base.ALL_ELEMENTS, target['column_index']), {'semantic_types': semantic_types})
            dataset.metadata = dataset.metadata.remove_semantic_type((target['resource_id'], 
            metadata_base.ALL_ELEMENTS, target['column_index']),'https://metadata.datadrivendiscovery.org/types/Attribute',)

    return dataset

def add_privileged_columns_metadata(dataset: 'Dataset', problem_doc: 'Metadata'):
    for privileged_data in problem_doc.get('inputs')[0].get('privileged_data', []):
        dataset.metadata = dataset.metadata.add_semantic_type((privileged_data['resource_id'], metadata_base.ALL_ELEMENTS, privileged_data['column_index']),'https://metadata.datadrivendiscovery.org/types/PrivilegedData',)

    return dataset

def add_target_metadata(dataset, targets):
    for target in targets:
        semantic_types = list(dataset.metadata.query((target.resource_id, metadata_base.ALL_ELEMENTS, target.column_index)).get('semantic_types', []))
        if 'https://metadata.datadrivendiscovery.org/types/Target' not in semantic_types:
            semantic_types.append('https://metadata.datadrivendiscovery.org/types/Target')
            dataset.metadata = dataset.metadata.update((target.resource_id, metadata_base.ALL_ELEMENTS, target.column_index), {'semantic_types': semantic_types})
        if 'https://metadata.datadrivendiscovery.org/types/TrueTarget' not in semantic_types:
            semantic_types.append('https://metadata.datadrivendiscovery.org/types/TrueTarget')
            dataset.metadata = dataset.metadata.update((target.resource_id, metadata_base.ALL_ELEMENTS, target.column_index), {'semantic_types': semantic_types})
        dataset.metadata = dataset.metadata.remove_semantic_type((target.resource_id,
        metadata_base.ALL_ELEMENTS, target.column_index),'https://metadata.datadrivendiscovery.org/types/Attribute',)

    return dataset

def add_privileged_metadata(dataset: 'Dataset', privileged_data):
    for data in privileged_data:
        dataset.metadata = dataset.metadata.add_semantic_type((data.resource_id, metadata_base.ALL_ELEMENTS, data.column_index),'https://metadata.datadrivendiscovery.org/types/PrivilegedData',)

    return dataset

def get_task(names):
    tasks = ['SEMISUPERVISED', 'OBJECTDETECTION', 'FORECASTING', 'GRAPHMATCHING', 'VERTEXNOMINATION', 'VERTEXCLASSIFICATION', \
             'COMMUNITYDETECTION', 'LINKPREDICTION', 'COLLABORATIVEFILTERING', 'CLUSTERING', 'REMOTESENSING', \
             'CLASSIFICATION', 'REGRESSION']
    for t in tasks:
        if t in names:
            if t == 'LINKPREDICTION':
                if 'TIMESERIES' in names:
                    return 'LINKPREDICTIONTIMESERIES'
            return t
    return None
        
def get_task_name(keywords):
    names = get_task_list(keywords)
    return get_task(names)

def get_task_list(keywords):
    names = []
    for k in keywords:
        name = k.upper()
        names.append(name)
    return names
 
def load_data_problem(inputdir, problempath):
    print("Reading ", inputdir)
    print("Reading ", problempath)

    with open(problempath) as file:
        problem_schema =  json.load(file)

    datasetId = problempath[:-29]
    dataset_schema = datasetId + "dataset_TRAIN/datasetDoc.json"
    problem_doc_metadata = Metadata(problem_schema)
    dataset_uri = 'file://{dataset_uri}'.format(dataset_uri=dataset_schema)
    dataset = D3MDatasetLoader().load(dataset_uri)

    problem_description = problem.parse_problem_description(problempath)
    dataset = add_target_columns_metadata(dataset, problem_description)
    dataset = add_privileged_columns_metadata(dataset, problem_description)
    taskname = get_task_name(problem_doc_metadata.query(())['about']['taskKeywords'])
    metric = problem_doc_metadata.query(())['inputs']['performanceMetrics'][0]['metric']
    posLabel = None
    if metric == "f1":
        posLabel = problem_doc_metadata.query(())['inputs']['performanceMetrics'][0]['posLabel']

    # Read the data augmentation
    keywords = getAugmentation_keywords(problem_doc_metadata)
  
    return (dataset, taskname, problem_description, metric, posLabel, keywords)

def load_testdata(problempath):
    datasetId = problempath[:-35]
    print("Id = ", datasetId)
    dataset_schema = datasetId + "/TEST/dataset_TEST/datasetDoc.json"
    dataset_uri = 'file://{dataset_uri}'.format(dataset_uri=dataset_schema)
    dataset = D3MDatasetLoader().load(dataset_uri)
    return dataset

def getAugmentation_keywords(problem_doc_metadata):
    keywords = None
    if "dataAugmentation" in problem_doc_metadata.query(()):
        keywords = problem_doc_metadata.query(())["dataAugmentation"]
    return keywords

def get_pipeline(dirname, pipeline_name):
    newdirname = dirname + "/" + pipeline_name
    filename = pipeline_name.split("_")[0]
    f = newdirname + "/" + filename + ".dump"

    solution = pickle.load(open(f, 'rb'))
    return solution

def write_solution(solution, dirname):
    rank = str(solution.rank)
    supporting_dirname = dirname + "/" + solution.id + "_" + rank
    if not os.path.exists(supporting_dirname):
        os.makedirs(supporting_dirname)
    output = open(supporting_dirname+"/"+solution.id+".dump", "wb")
    pickle.dump(solution, output)
    output.close()

def initialize_for_search(outputDir):
    name = outputDir
    if not os.path.exists(name):
        os.makedirs(name)

    dirNames = [outputDir+"/executables", outputDir+"/predictions", outputDir+"/pipelines_searched", outputDir+"/pipelines_scored", \
                outputDir+"/pipelines_ranked", outputDir+"/pipeline_runs", outputDir+"/subpipelines", outputDir+"/additional_inputs",
                outputDir+"/training_predictions"]
   
    for name in dirNames: 
        if not os.path.exists(name):
           os.makedirs(name)

def write_predictions(predictions, dirname, request_id):
    directory = dirname + "/" + str(request_id)
    if not os.path.exists(directory):
        os.makedirs(directory)

    outputFilePath = directory + "/predictions.csv"
    with open(outputFilePath, 'w') as outputFile:
        predictions.to_csv(outputFile, header=True, index=False)
    return outputFilePath
  
def write_feature_importances(feature_importances, dirname, request_id):
    directory = dirname + "/" + str(request_id)
    if not os.path.exists(directory):
        os.makedirs(directory)

    outputFilePath = directory + "/feature_importances.csv"
    with open(outputFilePath, 'w') as outputFile:
        feature_importances.to_csv(outputFile, header=True, index=False)
    return outputFilePath
 
def write_pipeline_json(solution, primitives, solution_dict, dirName, subpipeline_dirName, rank=None, train_score=None):
    solution.write_pipeline_json(primitives, solution_dict, dirName, subpipeline_dirName, rank, train_score) 

def write_rank_file(solution, rank, dirName):
    outputFilePath = dirName + "/" + solution.id + ".rank"
    with open(outputFilePath, 'w') as outputFile:
        outputFile.write(str(rank))
    
def write_pipeline_yaml(solution, dirname, dataset, problem_description):
    run_id = str(uuid.uuid4())
    filename = dirname + "/" + run_id + ".yaml"
    solution.write_pipeline_run(problem_description, dataset, filename)

def write_pipeline_executable(solution, dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    shell_script = '#!/bin/bash\n python ./autonml/main.py test\n'
    filename = dirname + "/" + solution.id + "_" + str(solution.rank) + ".sh"
    with open(filename, 'w') as f:
        f.write(shell_script)
    os.chmod(filename, 0o755)

def invert_metric(metric_type):
    min_metrics = set()
    min_metrics.add("MEAN_SQUARED_ERROR")
    min_metrics.add("ROOT_MEAN_SQUARED_ERROR")
    min_metrics.add("MEAN_ABSOLUTE_ERROR")
    min_metrics.add("LOSS")
    min_metrics.add("HAMMING_LOSS")
    if metric_type in min_metrics:
        return True
    return False

def get_distil_metric_name(metric_type):
    metric = 'accuracy'
    if metric_type == "MEAN_SQUARED_ERROR" or metric_type == "ROOT_MEAN_SQUARED_ERROR" or metric_type == "MEAN_ABSOLUTE_ERROR":
        metric = 'meanSquaredError'
    elif metric_type == "MEAN_ABSOLUTE_ERROR":
        metric = 'meanAbsoluteError'
    elif metric_type == "ACCURACY":
        metric = 'accuracy'
    elif metric_type == "F1_MACRO" or metric_type == "F1_MICRO" or metric_type == "F1":
        metric_type == 'f1Macro'
    return metric





# File: default_hyperparams.py 
# Author(s): Saswati Ray
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

__version__ = "0.1.0"

import os, json
import pandas as pd
import numpy as np

multimedia_threshold = 1200

def set_default_hyperparameters(path, taskname):
    """     
    Retrieve the default hyperparameters to be set for a primitive.
    Parameters     ---------     
    path: Python path of a primitive
    """

    hyperparams = None
    # Set hyperparameters for specific primitives
    if path == 'd3m.primitives.data_cleaning.imputer.SKlearn':
        hyperparams = {}
        hyperparams['use_semantic_types'] = True
        hyperparams['return_result'] = 'replace'
        hyperparams['strategy'] = 'median'
        hyperparams['error_on_no_input'] = False

    if path == 'd3m.primitives.data_transformation.one_hot_encoder.SKlearn':
        hyperparams = {}
        hyperparams['use_semantic_types'] = True
        hyperparams['return_result'] = 'replace'
        hyperparams['handle_unknown'] = 'ignore'

    if path == 'd3m.primitives.data_cleaning.robust_scaler.SKlearn':
        hyperparams = {}
        hyperparams['return_result'] = 'replace'

    if path == 'd3m.primitives.feature_construction.corex_text.DSBOX':
        hyperparams = {}
        hyperparams['threshold'] = 500

    if 'conditioner' in path:
        hyperparams = {}
        hyperparams['ensure_numeric'] = True
        hyperparams['maximum_expansion'] = 30

    if path == 'd3m.primitives.clustering.k_means.Fastlvm':
        hyperparams = {}
        hyperparams['k'] = 100

    if 'adjacency_spectral_embedding.JHU' in path:
        hyperparams = {}
        hyperparams['max_dimension'] = 5
        hyperparams['use_attributes'] = True
        if 'LINK' in taskname:
            hyperparams['max_dimension'] = 2
            hyperparams['use_attributes'] = False
            hyperparams['which_elbow'] = 1

    if 'splitter' in path:
        hyperparams = {}
        if taskname == 'IMAGE' or taskname == 'IMAGE2' or taskname == 'AUDIO' or taskname == 'VIDEO':
            hyperparams['threshold_row_length'] = multimedia_threshold
        else:
            hyperparams['threshold_row_length'] = 100000

    if path == 'd3m.primitives.link_prediction.link_prediction.DistilLinkPrediction':
        hyperparams = {}
        hyperparams['metric'] = 'accuracy'

    if path == 'd3m.primitives.vertex_nomination.seeded_graph_matching.DistilVertexNomination':
        hyperparams = {}
        hyperparams['metric'] = 'accuracy'

    if path == 'd3m.primitives.graph_matching.seeded_graph_matching.DistilSeededGraphMatcher':
        hyperparams = {}
        hyperparams['metric'] = 'accuracy'

    if 'PCA' in path:
        hyperparams = {}
        hyperparams['n_components'] = 10

    if path == 'd3m.primitives.data_transformation.image_reader.Common':
        hyperparams = {}
        hyperparams['use_columns'] = [0,1]
        hyperparams['return_result'] = 'replace'

    if path == 'd3m.primitives.data_transformation.text_reader.Common':
        hyperparams = {}
        hyperparams['return_result'] = 'replace'

    if path == 'd3m.primitives.time_series_forecasting.arima.DSBOX':
        hyperparams = {}
        hyperparams['take_log'] = False

    if path == 'd3m.primitives.time_series_forecasting.lstm.DeepAR':
        hyperparams = {}
        hyperparams['epochs'] = 3

    if path == 'd3m.primitives.graph_clustering.gaussian_clustering.JHU':
        hyperparams = {}
        hyperparams['max_clusters'] = 10 

    if path == 'd3m.primitives.time_series_forecasting.esrnn.RNN':
        hyperparams = {}
        hyperparams['auto_tune'] = True
        hyperparams['output_size'] = 60

    if path == 'd3m.primitives.time_series_forecasting.nbeats.DeepNeuralNetwork':
        hyperparams = {}
        hyperparams['loss'] = "MSE"
        hyperparams['seasonality'] = 24
        hyperparams['lr_decay'] = 0.5
        hyperparams['window_sampling_limit_multiplier'] = 200
        hyperparams['batch_size'] = 8

    if 'NBEATS' in path:
        hyperparams = {}
        hyperparams['prediction_length'] = 21
        hyperparams['num_context_lengths'] = 1
        hyperparams['output_mean'] = False

    if 'satellite_image_loader' in path:
        hyperparams = {}
        hyperparams['return_result'] = 'replace'

    if 'tfidf' in path:
        hyperparams = {}
        hyperparams['return_result'] = 'replace'
        hyperparams['use_semantic_types'] = True
        hyperparams['min_df'] = 0.025
        hyperparams['max_df'] = 0.85
        hyperparams['stop_words'] = 'english'

    if 'pca' in path:
        hyperparams = {}
        hyperparams['n_components'] = 10
        hyperparams['return_result'] = 'replace'

    if 'remote_sensing' in path:
        hyperparams = {}
        hyperparams['batch_size'] = 128
        hyperparams['pool_features'] = False

    if 'gcn_mixhop' in path:
        hyperparams = {}
        hyperparams['epochs'] = 200

    return hyperparams
